# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import pprint
import re
import sys
import os

import unittest

import pytest

from hgvs.exceptions import HGVSError, HGVSDataNotAvailableError, HGVSParseError, HGVSInvalidVariantError
from hgvs.enums import Datum
import hgvs.assemblymapper
import hgvs.dataproviders.uta
import hgvs.normalizer
import hgvs.parser
import hgvs.sequencevariant
import hgvs.alignmentmapper
import hgvs.validator
import hgvs.variantmapper
from support import CACHE


@pytest.mark.issues
class Test_Issues(unittest.TestCase):
    def setUp(self):
        self.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE)
        self.vm = hgvs.variantmapper.VariantMapper(self.hdp, replace_reference=False)
        self.vm_rr = hgvs.variantmapper.VariantMapper(self.hdp, replace_reference=True)
        self.hp = hgvs.parser.Parser()
        self.hn = hgvs.normalizer.Normalizer(self.hdp)
        self.hv = hgvs.validator.IntrinsicValidator()
        self.am37 = hgvs.assemblymapper.AssemblyMapper(
            self.hdp, replace_reference=True, assembly_name='GRCh37', alt_aln_method='splign')
        self.am38 = hgvs.assemblymapper.AssemblyMapper(
            self.hdp, replace_reference=True, assembly_name='GRCh38', alt_aln_method='splign')
        self.vn = hgvs.normalizer.Normalizer(self.hdp, shuffle_direction=3, cross_boundaries=True)

    def test_307_validator_rejects_valid_interval(self):
        # https://github.com/biocommons/hgvs/issues/307/
        # before fix, raises error; after fix, should pass
        self.hv.validate(self.hp.parse_hgvs_variant("NM_002858.3:c.1903-573_*1108del"))

    def test_314_parsing_identity_variant(self):
        v = self.hp.parse_hgvs_variant("NM_206933.2:c.6317=")
        self.assertEqual(str(v), "NM_206933.2:c.6317=")

        v = self.hp.parse_hgvs_variant("NM_206933.2:c.6317C=")
        self.assertEqual(str(v), "NM_206933.2:c.6317=")

    def test_316_provide_ttog_and_gtot_methods(self):
        h_c_g, h_c_t = "NC_000023.10:g.152864509_152864510insA", "NM_152274.3:c.21_22insT"
        h_nc_g, h_nc_t = "NC_000001.10:g.12776161G>A", "NR_111984.1:n.44G>A"

        v_c_g = self.hp.parse_hgvs_variant(h_c_g)
        v_c_t = self.hp.parse_hgvs_variant(h_c_t)
        v_nc_g = self.hp.parse_hgvs_variant(h_nc_g)
        v_nc_t = self.hp.parse_hgvs_variant(h_nc_t)

        self.assertEqual(h_c_t, str(self.am37.g_to_t(v_c_g, tx_ac=v_c_t.ac)))
        self.assertEqual(h_c_g, str(self.am37.t_to_g(v_c_t)))
        self.assertEqual(h_nc_t, str(self.am37.g_to_t(v_nc_g, tx_ac=v_nc_t.ac)))
        self.assertEqual(h_nc_g, str(self.am37.t_to_g(v_nc_t)))

    def test_317_improve_p_repr_of_syn_variants(self):
        # from original issue:
        v317 = self.hp.parse_hgvs_variant("NM_000059.3:c.7791A>G")
        self.assertEqual(str("NP_000050.2:p.(Lys2597=)"), str(self.am37.c_to_p(v317)))

    #def test_370_handle_multicodon_syn_variants(self):
    #    # Verify behavior of syn MNVs
    #    # NM_000059.3:
    #    #   :   3   6   9  12  15  18  21  24  27  30  33  36  39  42
    #    # 1 : ATG CCT ATT GGA TCC AAA GAG AGG CCA ACA TTT TTT GAA ATT
    #    #       M   P   I   G   W   F   E   R   P   T
    #    #                               ^      v21 (snv)
    #    #                                 ^    v22 (snv)
    #    #                                 ^^^  v2224 (mnv, 1 codon)
    #    #                               ^ ^    v2122 (mnv, 2 codons)
    #    # Glu (E): GA[AG]
    #    # Arg (R): AG[AG],  CG[UCAG]
    #    #
    #    # All of the following are synonymous changes:
    #    v21 = self.hp.parse_hgvs_variant("NM_000059.3:c.21G>A")  # GAG -> GAA (E)
    #    v22 = self.hp.parse_hgvs_variant("NM_000059.3:c.22A>C")  # AGG -> CGA (R)
    #    v2224 = self.hp.parse_hgvs_variant("NM_000059.3:c.22_24delinsCGA") # AGG -> CGA (R)
    #    v2122 = self.hp.parse_hgvs_variant("NM_000059.3:c.21_22delinsAC")  # GAGAGG -> GAACAG (ER)
    #
    #    self.assertEqual(b"NP_000050.2:p.(Glu7=)", str(self.am37.c_to_p(v21)))
    #    self.assertEqual(b"NP_000050.2:p.(Arg8=)", str(self.am37.c_to_p(v22)))
    #    self.assertEqual(b"NP_000050.2:p.(Arg8=)", str(self.am37.c_to_p(v2224)))
    #    self.assertEqual(b"NP_000050.2:p.(Glu7_Arg8=)", str(self.am37.c_to_p(v2122)))

    def test_322_raise_exception_when_mapping_bogus_variant(self):
        v = self.hp.parse_hgvs_variant("chrX:g.71684476delTGGAGinsAC")
        with self.assertRaises(HGVSInvalidVariantError):
            self.am37.g_to_c(v, "NM_018486.2")

    def test_324_error_normalizing_simple_inversion(self):
        v = self.hp.parse_hgvs_variant("NM_000535.5:c.1673_1674inv")
        vn = self.vn.normalize(v)
        self.assertEqual(str(vn), "NM_000535.5:c.1673_1674inv")    # no change

        vg = self.am37.c_to_g(v)
        v2 = self.am37.g_to_c(vg, tx_ac=v.ac)
        self.assertEqual(str(v), str(v2))    # no change after roundtrip

    def test_325_3p_UTR_treated_as_intronic(self):
        v = self.hp.parse_hgvs_variant("NM_000023.2:c.*6C>T")
        self.assertTrue(self.hv.validate(v))

    def test_326_handle_variants_in_par(self):
        "For a variant in the PAR, allow user user choose preference"
        hgvs_c = "NM_000451.3:c.584G>A"
        var_c = self.hp.parse_hgvs_variant(hgvs_c)

        self.am37.in_par_assume = None
        with self.assertRaises(HGVSError):
            var_g = self.am37.c_to_g(var_c)

        self.am37.in_par_assume = 'X'
        self.assertEqual(self.am37.c_to_g(var_c).ac, "NC_000023.10")

        self.am37.in_par_assume = 'Y'
        self.assertEqual(self.am37.c_to_g(var_c).ac, "NC_000024.9")

        self.am37.in_par_assume = None

    def test_330_incorrect_end_datum_post_ter(self):
        # https://github.com/biocommons/hgvs/issues/330/
        # In a variant like NM_004006.2:c.*87_91del, the interval
        # start and end are parsed independently. The * binds to
        # start but not end, causing end to have an incorrect datum.
        v = self.hp.parse_hgvs_variant("NM_004006.2:c.*87_91del")
        self.assertEqual(v.posedit.pos.start.datum, Datum.CDS_END)
        self.assertEqual(v.posedit.pos.end.datum, Datum.CDS_END)

    def test_334_delins_normalization(self):
        # also tests 335 re: inv including sequence (e.g., NC_000009.11:g.36233991_36233992invCA)
        v = self.hp.parse_hgvs_variant("NC_000009.11:g.36233991_36233992delCAinsAC")
        vn = self.vn.normalize(v)
        self.assertEqual(str(vn), "NC_000009.11:g.36233991_36233992delinsAC")

        v = self.hp.parse_hgvs_variant("NM_000535.5:c.1673_1674delCCinsGG")
        vn = self.vn.normalize(v)
        self.assertEqual(str(vn), "NM_000535.5:c.1673_1674inv")

        v = self.hp.parse_hgvs_variant("NM_000535.5:c.1_3delATGinsCAT")
        vn = self.vn.normalize(v)
        self.assertEqual(str(vn), "NM_000535.5:c.1_3inv")

    def test_346_reject_partial_alignments(self):
        # hgvs-346: verify that alignment data covers full-length transcript
        with self.assertRaises(HGVSDataNotAvailableError):
            hgvs.alignmentmapper.AlignmentMapper(
                self.hdp, tx_ac="NM_001290223.1", alt_ac="NC_000010.10", alt_aln_method="splign")

    def test_349_incorrect_normalization_insGG_to_dup2(self):
        # With dupN support, NM_005877.4:c.1104_1105insGG would
        # normalize to NM_005877.4:c.1104_1105dup2 (which is
        # different).
        original_var = "NM_005877.4:c.1104_1105insGG"
        self.assertEqual(original_var, str(
            self.hn.normalize(self.hp.parse_c_variant(original_var))))

    def test_379_move_replace_reference_to_variantmapper(self):
        # replace_reference code was in am, not vm. That meant that using vm directly
        # resulted in variants that were not reference corrected.
        g_var = self.hp.parse_hgvs_variant("NC_000006.11:g.44275011T=")
        c_var = self.hp.parse_hgvs_variant(
            "NM_020745.3:c.1015G>A")    # correct projection with ref replacement
        self.assertEqual(c_var, self.am37.g_to_c(g_var, "NM_020745.3"))    # previously okay
        self.assertEqual(c_var, self.vm_rr.g_to_c(g_var, "NM_020745.3"))    # previously wrong

    def test_381_c_to_p_error_with_del_variants(self):
        hgvs_c = "NM_000302.3:c.1594_1596del"
        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        self.am37.c_to_p(var_c)    # raises exception before fixing

    def test_393_posedit_for_unknown_p_effect(self):
        hgvs_c = "NM_001330368.1:c.641-3353C>A"
        hgvs_p = "NP_001317297.1:p.?"
        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        translated_var_p = self.am37.c_to_p(var_c)
        parsed_var_p = self.hp.parse_hgvs_variant(hgvs_p)
        self.assertIsNone(parsed_var_p.posedit)
        self.assertIsNone(translated_var_p.posedit)
        self.assertEqual(hgvs_p, str(translated_var_p))



# -*- coding: utf-8 -*-
"""The hgvs package uses a single, package-wide configuration instance
to control package behavior.  The hgvs.config module provides that
configuration instance, via the `hgvs.global_config` variable.

You should not import hgvs.config directly.

Config are read from an ini-format file.  `hgvs.config` implements a
thin wrapper on the ConfigParser instance in order to provide
*attribute* based lookups (rather than key). It also returns
heuristically typed values (e.g., "True" becomes True). 

Although keys are settable, they are stringified on setting and
type-inferred on getting, which means that round-tripping works only
for str, int, and boolean.

>>> import hgvs.config

.. data:: hgvs.config.global_config

   Package-wide ("global") configuration, initialized with package
   defaults.  Setting configuration in this object will change global
   behavior of the hgvs package.

   global_config, an instance of :ref:``hgvs.config.Config``, supports
   reading ini-like files that updates

"""

from __future__ import absolute_import, division, print_function, unicode_literals

from configparser import ConfigParser, ExtendedInterpolation
from copy import copy
import logging
from pkg_resources import resource_stream
import re

logger = logging.getLogger(__name__)


class Config(object):
    """provides an attribute-based lookup of configparser sections and
    settings.

    """

    def __init__(self, extended_interpolation=True):
        if extended_interpolation:
            cp = ConfigParser(interpolation=ExtendedInterpolation())
        else:
            cp = ConfigParser()
        cp.optionxform = _name_xform
        self._cp = cp

    def read_stream(self, flo):
        """read configuration from ini-formatted file-like object

        """
        self._cp.read_string(flo.read().decode('ascii'))

    def __copy__(self):
        new_config = Config.__new__(Config)
        new_config._cp = object.__getattribute__(self, '_cp')
        return new_config

    def __dir__(self):
        return list(self._cp.keys())

    def __getattr__(self, k):
        # Work around PyCharm bug https://youtrack.jetbrains.com/issue/PY-4213
        if k == "_cp":
            return
        try:
            return ConfigGroup(self._cp[k])
        except KeyError:
            raise AttributeError(k)

    __getitem__ = __getattr__


class ConfigGroup(object):
    def __init__(self, section):
        self.__dict__["_section"] = section

    def __dir__(self):
        return list(self.__dict__["_section"].keys())

    def __getattr__(self, k):
        return _val_xform(self.__dict__["_section"][k])

    __getitem__ = __getattr__

    def __setattr__(self, k, v):
        logger.info(str(self.__class__.__name__) + ".__setattr__({k}, ...)".format(k=k))
        self.__dict__["_section"][k] = str(v)

    __setitem__ = __setattr__


def _name_xform(o):
    """transform names to lowercase, without symbols (except underscore)
    Any chars other than alphanumeric are converted to an underscore
    """
    return re.sub(r"\W", "_", o.lower())


def _val_xform(v):
    if v == "True":
        return True
    if v == "False":
        return False
    if v == "None":
        return None
    try:
        return int(v)
    except ValueError:
        pass
    return v


_default_config = Config()
_default_config.read_stream(resource_stream(__name__, "_data/defaults.ini"))

global_config = copy(_default_config)



#!/usr/bin/env python

import hgvs
import hgvs.parser
from tabulate import tabulate
from six.moves import map

hp = hgvs.parser.Parser()

variants = [
    "NM_01.2:c.1_2del",
    "NM_01.2:c.1_2del2",
    "NM_01.2:c.1_2delAA",
    "NM_01.2:c.1_2insTTT",
    "NM_01.2:c.1_2delinsTTT",
    "NM_01.2:c.1_2del2insTTT",
    "NM_01.2:c.1_2delAAinsTTT",
    "NM_01.2:c.1_2delAAinsAA",
    "NM_01.2:c.1_1delAinsA",
    "NM_01.2:c.1A>A",
    "NM_01.2:c.1_1delAinsT",
    "NM_01.2:c.1A>T",
    "NM_01.2:c.1=",
    "NM_01.2:c.1A=",
    "NM_01.2:c.1AA=",
]


headers = "hgvs ref alt etype delta".split()
def gen1(h):
    v = hp.parse_hgvs_variant(h)
    return [h, v.posedit.edit.ref, v.posedit.edit.alt, v.posedit.edit.type, v.posedit.length_change()]


rows = [list(map(str, gen1(h))) for h in variants]

print(tabulate(rows, headers=headers)) #, tablefmt="pipe"))



# -*- coding: utf-8 -*-
"""Provides coordinate (not variant) mapping operations between
genomic (g), non-coding (n), cds (c), and protein (p) coordinates.

"""

from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from bioutils.coordinates import strand_int_to_pm

import hgvs.intervalmapper
import hgvs.location

from hgvs.exceptions import HGVSError, HGVSUsageError, HGVSDataNotAvailableError
from hgvs.utils import build_tx_cigar
from hgvs.enums import Datum
from six.moves import range

_logger = logging.getLogger(__name__)
_logger.warning("This module is deprecated and will be removed in a future release")


class TranscriptMapper(object):
    """Provides coordinate (not variant) mapping operations between
    genomic (g), non-coding (n), cds (c), and protein (p) coordinates.
    All coordinates are 1-based inclusive, per the HGVS
    recommendations.  All methods take :class:`hgvs.location.Interval`
    objects.

    :param hdp: HGVS Data Provider Interface-compliant instance (see :class:`hgvs.dataproviders.interface.Interface`)
    :param str tx_ac: string representing transcript accession (e.g., NM_000551.2)
    :param str alt_ac: string representing the reference sequence accession (e.g., NM_000551.3)
    :param str alt_aln_method: string representing the alignment method; valid values depend on data source

    """

    def __init__(self, hdp, tx_ac, alt_ac, alt_aln_method):
        self.hdp = hdp
        self.tx_ac = tx_ac
        self.alt_ac = alt_ac
        self.alt_aln_method = alt_aln_method
        if self.alt_aln_method != "transcript":
            self.tx_info = hdp.get_tx_info(self.tx_ac, self.alt_ac, self.alt_aln_method)
            if self.tx_info is None:
                raise HGVSDataNotAvailableError(
                    "TranscriptMapper(tx_ac={self.tx_ac}, "
                    "alt_ac={self.alt_ac}, alt_aln_method={self.alt_aln_method}): "
                    "No transcript info".format(self=self))

            self.tx_exons = hdp.get_tx_exons(self.tx_ac, self.alt_ac, self.alt_aln_method)
            if self.tx_exons is None:
                raise HGVSDataNotAvailableError(
                    "TranscriptMapper(tx_ac={self.tx_ac}, "
                    "alt_ac={self.alt_ac}, alt_aln_method={self.alt_aln_method}): "
                    "No transcript exons".format(self=self))

            # hgvs-386: An assumption when building the cigar string
            # is that exons are adjacent. Assert that here.
            tx_exons = sorted(self.tx_exons, key=lambda e: e["ord"])
            for i in range(1, len(tx_exons)):
                if tx_exons[i - 1]["tx_end_i"] != tx_exons[i]["tx_start_i"]:
                    raise HGVSDataNotAvailableError(
                        "TranscriptMapper(tx_ac={self.tx_ac}, "
                        "alt_ac={self.alt_ac}, alt_aln_method={self.alt_aln_method}): "
                        "Exons {a} and {b} are not adjacent".format(self=self, a=i, b=i + 1))

            self.strand = self.tx_exons[0]["alt_strand"]
            self.cds_start_i = self.tx_info["cds_start_i"]
            self.cds_end_i = self.tx_info["cds_end_i"]
            self.gc_offset = self.tx_exons[0]["alt_start_i"]
            self.cigar = build_tx_cigar(self.tx_exons, self.strand)
            self.im = hgvs.intervalmapper.IntervalMapper.from_cigar(self.cigar)
            self.tgt_len = self.im.tgt_len
        else:
            # this covers the identity cases n <-> c
            self.tx_identity_info = hdp.get_tx_identity_info(self.tx_ac)
            if self.tx_identity_info is None:
                raise HGVSError("TranscriptMapper(tx_ac={self.tx_ac}, "
                                "alt_ac={self.alt_ac}, alt_aln_method={self.alt_aln_method}): "
                                "No transcript identity info".format(self=self))
            self.cds_start_i = self.tx_identity_info["cds_start_i"]
            self.cds_end_i = self.tx_identity_info["cds_end_i"]
            self.tgt_len = sum(self.tx_identity_info["lengths"])

        assert not (
            (self.cds_start_i is None) ^
            (self.cds_end_i is None)), "CDS start and end must both be defined or neither defined"

    def __str__(self):
        return "{self.__class__.__name__}: {self.tx_ac} ~ {self.alt_ac} ~ {self.alt_aln_method); " \
               "{strand_pm} strand; {n_exons} exons; offset={self.gc_offset}".format(
                   self=self, n_exons=len(self.tx_exons), strand_pm=strand_int_to_pm(self.strand))

    @property
    def is_coding_transcript(self):
        if ((self.tx_info["cds_start_i"] is not None) ^ (self.tx_info["cds_end_i"] is not None)):
            raise HGVSError("{self.tx_ac}: CDS start_i and end_i"
                            " must be both defined or both undefined".format(self=self))
        return self.tx_info["cds_start_i"] is not None

    def g_to_n(self, g_interval):
        """convert a genomic (g.) interval to a transcript cDNA (n.) interval"""

        # This code is extremely convoluted. To begin with, it
        # confuses interbase intervals for a *single* hgvs position
        # with intervals of *two* hgvs positions.  This is the origin
        # of bug #205 (now closed). Although the bug is itself fixed,
        # the mappers really need an overhaul. See #208.

        def _hgvs_offset(g_position, grs, gre, strand):
            """ Calculates the HGVS coordinate offset from a given genomic position"""
            if g_position == grs or g_position == gre:
                return 0
            mid = (grs + gre) / 2
            if g_position < mid or (g_position == mid and strand == 1):
                offset = g_position - grs
            else:
                offset = g_position - gre
            return offset

        def map_g_to_n_pos(pos):
            g_ci = _hgvs_coord_to_ci(pos, pos)
            # frs, fre = (f)orward (r)na (s)tart & (e)nd; forward w.r.t. genome
            frs, fre = self.im.map_ref_to_tgt(
                g_ci[0] - self.gc_offset, g_ci[1] - self.gc_offset, max_extent=False)
            if self.strand == -1:
                frs, fre = self.tgt_len - fre, self.tgt_len - frs
            # get BaseOffsetPosition information for n.
            if self.strand == 1:
                grs, gre = self.im.map_tgt_to_ref(frs, fre, max_extent=False)
            elif self.strand == -1:
                grs, gre = self.im.map_tgt_to_ref((self.tgt_len - fre), (self.tgt_len - frs),
                                                  max_extent=False)
            grs, gre = grs + self.gc_offset, gre + self.gc_offset
            # 0-width interval indicates an intron.  Need to calculate offsets but we"re are in ci coordinates
            # requires adding 1 strategically to get the HGVS position
            # (shift coordinates to 3' end of the ref nucleotide)

            if frs == fre:
                start_offset = _hgvs_offset(g_ci[0] + 1, grs, gre + 1, self.strand)
                end_offset = _hgvs_offset(g_ci[1], grs, gre + 1, self.strand)
            else:
                start_offset = _hgvs_offset(g_ci[0], grs, gre, self.strand)
                end_offset = _hgvs_offset(g_ci[1], grs, gre, self.strand)
            if self.strand == -1:
                start_offset, end_offset = self.strand * end_offset, self.strand * start_offset
            if start_offset > 0:
                frs -= 1
            if end_offset < 0:
                fre += 1
            start_base = _ci_to_hgvs_coord(frs, fre)[0]
            end_base = _ci_to_hgvs_coord(frs, fre)[1]
            return start_base, start_offset, end_base, end_offset

        start_bo = map_g_to_n_pos(g_interval.start.base)[0:2]
        end_bo = start_bo if g_interval.start.base == g_interval.end.base else map_g_to_n_pos(
            g_interval.end.base)[2:4]
        if self.strand == -1:
            start_bo, end_bo = end_bo, start_bo

        return hgvs.location.BaseOffsetInterval(
            start=hgvs.location.BaseOffsetPosition(
                base=start_bo[0], offset=start_bo[1], datum=Datum.SEQ_START),
            end=hgvs.location.BaseOffsetPosition(
                base=end_bo[0], offset=end_bo[1], datum=Datum.SEQ_START),
            uncertain=g_interval.uncertain)

    def n_to_g(self, n_interval):
        """convert a transcript cDNA (n.) interval to a genomic (g.) interval"""

        assert self.strand in [1, -1], "strand = " + str(self.strand) + "; must be 1 or -1"

        if self.strand == 1:
            frs, fre = _hgvs_coord_to_ci(n_interval.start.base, n_interval.end.base)
            start_offset, end_offset = n_interval.start.offset, n_interval.end.offset
        elif self.strand == -1:
            frs, fre = _hgvs_coord_to_ci(n_interval.start.base, n_interval.end.base)
            fre, frs = self.tgt_len - frs, self.tgt_len - fre
            start_offset, end_offset = self.strand * n_interval.end.offset, self.strand * n_interval.start.offset

        # returns the genomic range start (grs) and end (gre)
        grs, gre = self.im.map_tgt_to_ref(frs, fre, max_extent=False)
        grs, gre = grs + self.gc_offset, gre + self.gc_offset
        gs, ge = grs + start_offset, gre + end_offset
        return hgvs.location.Interval(
            start=hgvs.location.SimplePosition(
                _ci_to_hgvs_coord(gs, ge)[0], uncertain=n_interval.start.uncertain),
            end=hgvs.location.SimplePosition(
                _ci_to_hgvs_coord(gs, ge)[1], uncertain=n_interval.end.uncertain),
            uncertain=n_interval.uncertain)

    def n_to_c(self, n_interval):
        """convert a transcript cDNA (n.) interval to a transcript CDS (c.) interval"""

        if self.cds_start_i is None:    # cds_start_i defined iff cds_end_i defined; see assertion above
            raise HGVSUsageError(
                "CDS is undefined for {self.tx_ac}; cannot map to c. coordinate (non-coding transcript?)"
                .format(self=self))
        if n_interval.start.base <= 0 or n_interval.end.base > self.tgt_len:
            raise HGVSError("The given coordinate is outside the bounds of the reference sequence.")

        # start
        if n_interval.start.base <= self.cds_start_i:
            cs = n_interval.start.base - (self.cds_start_i + 1)
            cs_datum = Datum.CDS_START
        elif n_interval.start.base > self.cds_start_i and n_interval.start.base <= self.cds_end_i:
            cs = n_interval.start.base - self.cds_start_i
            cs_datum = Datum.CDS_START
        else:
            cs = n_interval.start.base - self.cds_end_i
            cs_datum = Datum.CDS_END
        # end
        if n_interval.end.base <= self.cds_start_i:
            ce = n_interval.end.base - (self.cds_start_i + 1)
            ce_datum = Datum.CDS_START
        elif n_interval.end.base > self.cds_start_i and n_interval.end.base <= self.cds_end_i:
            ce = n_interval.end.base - self.cds_start_i
            ce_datum = Datum.CDS_START
        else:
            ce = n_interval.end.base - self.cds_end_i
            ce_datum = Datum.CDS_END

        c_interval = hgvs.location.BaseOffsetInterval(
            start=hgvs.location.BaseOffsetPosition(
                base=cs, offset=n_interval.start.offset, datum=cs_datum),
            end=hgvs.location.BaseOffsetPosition(
                base=ce, offset=n_interval.end.offset, datum=ce_datum),
            uncertain=n_interval.uncertain)
        return c_interval

    def c_to_n(self, c_interval):
        """convert a transcript CDS (c.) interval to a transcript cDNA (n.) interval"""

        if self.cds_start_i is None:    # cds_start_i defined iff cds_end_i defined; see assertion above
            raise HGVSUsageError(
                "CDS is undefined for {self.tx_ac}; cannot map from c. coordinate (non-coding transcript?)"
                .format(self=self))

        # start
        if c_interval.start.datum == Datum.CDS_START and c_interval.start.base < 0:
            rs = c_interval.start.base + self.cds_start_i + 1
        elif c_interval.start.datum == Datum.CDS_START and c_interval.start.base > 0:
            rs = c_interval.start.base + self.cds_start_i
        elif c_interval.start.datum == Datum.CDS_END:
            rs = c_interval.start.base + self.cds_end_i
        # end
        if c_interval.end.datum == Datum.CDS_START and c_interval.end.base < 0:
            re = c_interval.end.base + self.cds_start_i + 1
        elif c_interval.end.datum == Datum.CDS_START and c_interval.end.base > 0:
            re = c_interval.end.base + self.cds_start_i
        elif c_interval.end.datum == Datum.CDS_END:
            re = c_interval.end.base + self.cds_end_i

        if rs <= 0 or re > self.tgt_len:
            raise HGVSError("The given coordinate is outside the bounds of the reference sequence.")

        n_interval = hgvs.location.BaseOffsetInterval(
            start=hgvs.location.BaseOffsetPosition(
                base=rs, offset=c_interval.start.offset, datum=Datum.SEQ_START),
            end=hgvs.location.BaseOffsetPosition(
                base=re, offset=c_interval.end.offset, datum=Datum.SEQ_START),
            uncertain=c_interval.uncertain)
        return n_interval

    def g_to_c(self, g_interval):
        """convert a genomic (g.) interval to a transcript CDS (c.) interval"""
        return self.n_to_c(self.g_to_n(g_interval))

    def c_to_g(self, c_interval):
        """convert a transcript CDS (c.) interval to a genomic (g.) interval"""
        return self.n_to_g(self.c_to_n(c_interval))


def _ci_to_hgvs_coord(s, e):
    """ Convert continuous interbase (right-open) coordinates (..,-2,-1,0,1,..) to
    discontinuous HGVS coordinates (..,-2,-1,1,2,..)
    """

    def _ci_to_hgvs(c):
        return c + 1 if c >= 0 else c

    return (None if s is None else _ci_to_hgvs(s), None if e is None else _ci_to_hgvs(e) - 1)


def _hgvs_coord_to_ci(s, e):
    """convert start,end interval in inclusive, discontinuous HGVS coordinates
    (..,-2,-1,1,2,..) to continuous interbase (right-open) coordinates
    (..,-2,-1,0,1,..)"""

    def _hgvs_to_ci(c):
        assert c != 0, "received CDS coordinate 0; expected ..,-2,-1,1,1,..."
        return c - 1 if c > 0 else c

    return (None if s is None else _hgvs_to_ci(s), None if e is None else _hgvs_to_ci(e) + 1)


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from enum import Enum


class OrderedEnum(Enum):
    def __ge__(self, other):
        assert self.__class__ is other.__class__, "OrderedEnum can only compare to OrderedEnum"
        return self.value >= other.value

    def __gt__(self, other):
        assert self.__class__ is other.__class__, "OrderedEnum can only compare to OrderedEnum"
        return self.value > other.value

    def __le__(self, other):
        assert self.__class__ is other.__class__, "OrderedEnum can only compare to OrderedEnum"
        return self.value <= other.value

    def __lt__(self, other):
        assert self.__class__ is other.__class__, "OrderedEnum can only compare to OrderedEnum"
        return self.value < other.value


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
"""Provides parser for HGVS strings and HGVS-related conceptual
components, such as intronic-offset coordiates

"""

from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import copy
import re

from pkg_resources import resource_filename

import bioutils.sequences
import ometa.runtime
import parsley

from hgvs.exceptions import HGVSParseError

# The following imports are referenced by fully-qualified name in the
# hgvs grammar.
import hgvs.enums
import hgvs.edit
import hgvs.hgvsposition
import hgvs.location
import hgvs.posedit
import hgvs.sequencevariant


class Parser(object):
    """Provides comprehensive parsing of HGVS varaint strings (*i.e.*,
    variants represented according to the Human Genome Variation
    Society recommendations) into Python representations.  The class
    wraps a Parsing Expression Grammar, exposing rules of that grammar
    as methods (prefixed with `parse_`) that parse an input string
    according to the rule.  The class exposes all rules, so that it's
    possible to parse both full variant representations as well as
    components, like so:

    >>> hp = Parser()
    >>> v = hp.parse_hgvs_variant("NM_01234.5:c.22+1A>T")
    >>> v
    SequenceVariant(ac=NM_01234.5, type=c, posedit=22+1A>T, gene=None)
    >>> v.posedit.pos
    BaseOffsetInterval(start=22+1, end=22+1, uncertain=False)
    >>> i = hp.parse_c_interval("22+1")
    >>> i
    BaseOffsetInterval(start=22+1, end=22+1, uncertain=False)

    The `parse_hgvs_variant` and `parse_c_interval` methods correspond
    to the `hgvs_variant` and `c_interval rules` in the grammar,
    respectively.

    As a convenience, the Parser provides the `parse` method as a
    shorthand for `parse_hgvs_variant`:
    >>> v = hp.parse("NM_01234.5:c.22+1A>T")
    >>> v
    SequenceVariant(ac=NM_01234.5, type=c, posedit=22+1A>T, gene=None)

    Because the methods are generated on-the-fly and depend on the
    grammar that is loaded at runtime, a full list of methods is not
    available in the documentation.  However, the list of
    rules/methods is available via the `rules` instance variable.

    A few notable methods are listed below:

    `parse_hgvs_variant()` parses any valid HGVS string supported by the grammar.

      >>> hp.parse_hgvs_variant("NM_01234.5:c.22+1A>T")
      SequenceVariant(ac=NM_01234.5, type=c, posedit=22+1A>T, gene=None)
      >>> hp.parse_hgvs_variant("NP_012345.6:p.Ala22Trp")
      SequenceVariant(ac=NP_012345.6, type=p, posedit=Ala22Trp, gene=None)

    The `hgvs_variant` rule iteratively attempts parsing using the
    major classes of HGVS variants. For slight improvements in
    efficiency, those rules may be invoked directly:

      >>> hp.parse_p_variant("NP_012345.6:p.Ala22Trp")
      SequenceVariant(ac=NP_012345.6, type=p, posedit=Ala22Trp, gene=None)

    Similarly, components of the underlying structure may be parsed
    directly as well:

      >>> hp.parse_c_posedit("22+1A>T")
      PosEdit(pos=22+1, edit=A>T, uncertain=False)
      >>> hp.parse_c_interval("22+1")
      BaseOffsetInterval(start=22+1, end=22+1, uncertain=False)

    """

    __default_grammar_fn = resource_filename(__name__, "_data/hgvs.pymeta")

    def __init__(self, grammar_fn=__default_grammar_fn, expose_all_rules=False):
        self._grammar_fn = grammar_fn
        self._grammar = parsley.makeGrammar(
            open(grammar_fn, "r").read(), {
                "hgvs": hgvs,
                "bioutils": bioutils,
                "copy": copy
            })
        self._logger = logging.getLogger(__name__)
        self._expose_rule_functions(expose_all_rules)

    def parse(self, v):
        """parse HGVS variant `v`, returning a SequenceVariant

        :param str v: an HGVS-formatted variant as a string
        :rtype: SequenceVariant

        """
        return self.parse_hgvs_variant(v)

    def _expose_rule_functions(self, expose_all_rules=False):
        """add parse functions for public grammar rules

        Defines a function for each public grammar rule, based on
        introspecting the grammar. For example, the `c_interval` rule
        is exposed as a method `parse_c_interval` and used like this::

          Parser.parse_c_interval('26+2_57-3') -> Interval(...)

        """

        def make_parse_rule_function(rule_name):
            "builds a wrapper function that parses a string with the specified rule"

            def rule_fxn(s):
                try:
                    return self._grammar(s).__getattr__(rule_name)()
                except ometa.runtime.ParseError as exc:
                    raise HGVSParseError("{s}: char {exc.position}: {reason}".format(
                        s=s, exc=exc, reason=exc.formatReason()))

            rule_fxn.__doc__ = "parse string s using `%s' rule" % rule_name
            return rule_fxn

        exposed_rule_re = re.compile(r"hgvs_(variant|position)|(c|g|m|n|p|r)"
                                     r"_(edit|hgvs_position|interval|pos|posedit|variant)")
        exposed_rules = [
            m.replace("rule_", "") for m in dir(self._grammar._grammarClass)
            if m.startswith("rule_")
        ]
        if not expose_all_rules:
            exposed_rules = [
                rule_name for rule_name in exposed_rules if exposed_rule_re.match(rule_name)
            ]
        for rule_name in exposed_rules:
            att_name = "parse_" + rule_name
            rule_fxn = make_parse_rule_function(rule_name)
            self.__setattr__(att_name, rule_fxn)
        self._logger.debug("Exposed {n} rules ({rules})".format(
            n=len(exposed_rules), rules=", ".join(exposed_rules)))


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
"""Tests uta postgresql client"""

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import re
import unittest

from hgvs.exceptions import HGVSDataNotAvailableError
import hgvs.dataproviders.uta
import hgvs.edit
import hgvs.location
import hgvs.posedit
import hgvs.variantmapper
import hgvs.sequencevariant
from support import CACHE


class UTA_Base(object):
    def test_get_acs_for_protein_seq(self):
        exp = ["NP_001005405.1", "MD5_8fc09b1d9a38a8c55176a0fa922df227"]
        s = """
        mgccgcsggc gsgcggcgsg sggcgsgcgg cgssccvpic cckpvcccvp acscsscgsc
        ggskggcgsc gsskggcgsc gcsqsncckp ccsssgcgsf ccqsscskpc ccqssccqss
        cckpcccqss ccqsscfkpc ccqssccvpv ccqcki
        """

        s = re.sub(r"\s+", "", s.upper())
        self.assertEqual(sorted(self.hdp.get_acs_for_protein_seq(s)), sorted(exp))

        exp = ["NP_071928.2", "MD5_ffb0d4adbd5e0b5d71678228b3696984"]
        s = """
        masetektha llqtcstesl isslglgafc lvadrllqfs tiqqndwlra lsdnavhcvi
        gmwswavvtg ikkktdfgei ilagflasvi dvdhfflags mslkaaltlp rrpflhcstv
        ipvvvltlkf tmhlfklkds wcflpwmlfi swtshhirdg irhglwicpf gktsplpfwl
        yviitsslph icsfvmyltg trqmmsskhg vridv
        """

        s = re.sub(r"\s+", "", s.upper())
        self.assertEqual(sorted(self.hdp.get_acs_for_protein_seq(s)), sorted(exp))

    def test_get_gene_info(self):
        gene_info = self.hdp.get_gene_info("VHL")
        self.assertEqual("VHL", gene_info["hgnc"])
        self.assertEqual("3p25.3", gene_info["maploc"])
        self.assertEqual(6, len(gene_info))

    def test_get_tx_exons(self):
        tx_exons = self.hdp.get_tx_exons("NM_000551.3", "NC_000003.11", "splign")
        self.assertEqual(3, len(tx_exons))

    def test_get_tx_exons_invalid_tx_ac(self):
        with self.assertRaises(HGVSDataNotAvailableError):
            self.hdp.get_tx_exons("NM_999999.9", "NC_000003.11", "splign")

    def test_get_tx_exons_invalid_alt_ac(self):
        with self.assertRaises(HGVSDataNotAvailableError):
            self.hdp.get_tx_exons("NM_000551.3", "NC_000999.9", "splign")

    def test_get_tx_exons_invalid_alt_aln_method(self):
        with self.assertRaises(HGVSDataNotAvailableError):
            self.hdp.get_tx_exons("NM_000551.3", "NC_000999.9", "best")

    def test_get_tx_for_gene(self):
        tig = self.hdp.get_tx_for_gene("VHL")
        self.assertEqual(14, len(tig))

    def test_get_tx_for_gene_invalid_gene(self):
        tig = self.hdp.get_tx_for_gene("GENE")
        self.assertEqual(0, len(tig))

    def test_get_tx_info(self):
        tx_info = self.hdp.get_tx_info("NM_000051.3", "AC_000143.1", "splign")
        self.assertEqual(385, tx_info["cds_start_i"])
        self.assertEqual(9556, tx_info["cds_end_i"])
        self.assertEqual("AC_000143.1", tx_info["alt_ac"])

    def test_get_tx_info_invalid_tx_ac(self):
        with self.assertRaises(HGVSDataNotAvailableError):
            self.hdp.get_tx_info("NM_999999.9", "AC_000143.1", "splign")

    def test_get_tx_mapping_options(self):
        tx_mapping_options = self.hdp.get_tx_mapping_options("NM_000551.3")
        self.assertIn(["NM_000551.3", "NC_000003.11", "splign"], tx_mapping_options)
        self.assertIn(["NM_000551.3", "NC_000003.11", "blat"], tx_mapping_options)

    def test_get_tx_mapping_options_invalid(self):
        tx_info_options = self.hdp.get_tx_mapping_options("NM_999999.9")
        self.assertEqual(tx_info_options, [])


class Test_hgvs_dataproviders_uta_UTA_default(unittest.TestCase, UTA_Base):
    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE)


class Test_hgvs_dataproviders_uta_UTA_default_with_pooling(unittest.TestCase, UTA_Base):
    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect(
            pooling=True, mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE)


class TestUTACache(Test_hgvs_dataproviders_uta_UTA_default):
    def _create_cdna_variant(self):
        start = hgvs.location.SimplePosition(118898437)
        end = hgvs.location.SimplePosition(118898437)
        iv = hgvs.location.Interval(start=start, end=end)
        edit = hgvs.edit.NARefAlt(ref="C", alt="T")
        posedit = hgvs.posedit.PosEdit(pos=iv, edit=edit)
        genomic_variant = hgvs.sequencevariant.SequenceVariant(
            ac="NC_000011.9",
            type="g",
            posedit=posedit,
        )
        variantmapper = hgvs.variantmapper.VariantMapper(self.hdp)
        return variantmapper.g_to_c(genomic_variant, "NM_001164277.1")

    def test_deterministic_cache_results(self):
        """
        Check that identical request to the UTA yields the same results.
        """
        var1 = self._create_cdna_variant()
        var2 = self._create_cdna_variant()
        self.assertEqual(str(var1), str(var2))


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os

import pytest

from hgvs.exceptions import HGVSError, HGVSInvalidVariantError
import hgvs.dataproviders.uta
import hgvs.parser
import hgvs.assemblymapper
from support import CACHE

tests_fn = "tests/data/proj-near-disc.tsv"


def read_tests(fn):
    """read tests from tsv file, return iterator of dicts"""
    fh = open(fn)
    for line in fh:
        line = line.strip()
        if line == "":
            continue
        if line.startswith("#"):
            continue
        dt, lt, var, exp = line.split()
        yield {"disc_type": dt, "loc_type": lt, "variant": var, "expected": exp}


hp = hgvs.parser.Parser()
hdp = hgvs.dataproviders.uta.connect(mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE)
# TODO: Use variantmapper instead of assemblymapper
am38 = hgvs.assemblymapper.AssemblyMapper(hdp, assembly_name='GRCh38', normalize=False)

tests = list(read_tests(tests_fn))
params = [(t["variant"], t["expected"], t["loc_type"] + " " + t["disc_type"]) for t in tests]


@pytest.mark.parametrize("variant,expected,description", params)
def test_projection_near_discrepancies(variant, expected, description):
    var_n = hp.parse_hgvs_variant(variant)
    var_g = am38.n_to_g(var_n)
    assert expected == str(var_g), description


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

#
# Tests for conversion of hgvs tags
#
import os
import unittest

import hgvs.variantmapper as variantmapper
import hgvs.parser

import support.mock_input_source as mock_input_data_source


class TestHgvsCToP(unittest.TestCase):

    fn = os.path.join(os.path.dirname(__file__), "data", "sanity_cp.tsv")
    _datasource = mock_input_data_source.MockInputSource(fn)
    _mapper = variantmapper.VariantMapper(_datasource, prevalidation_level="INTRINSIC")
    _parser = hgvs.parser.Parser()

    def test_silent(self):
        hgvsc = "NM_999999.1:c.6A>G"
        hgvsp_expected = "MOCK:p.(Lys2=)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_substitution(self):
        hgvsc = "NM_999999.1:c.6A>T"
        hgvsp_expected = "MOCK:p.(Lys2Asn)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_substitution_introduces_stop_codon(self):
        hgvsc = "NM_999996.1:c.8C>A"
        hgvsp_expected = "MOCK:p.(Ser3Ter)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_substitution_removes_stop_codon(self):
        hgvsc = "NM_999998.1:c.30G>T"
        hgvsp_expected = "MOCK:p.(Ter10TyrextTer3)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_insertion_no_frameshift(self):
        hgvsc = "NM_999999.1:c.6_7insGGG"
        hgvsp_expected = "MOCK:p.(Lys2_Ala3insGly)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_insertion_frameshift(self):
        hgvsc = "NM_999999.1:c.22_23insT"
        hgvsp_expected = "MOCK:p.(Ala8ValfsTer?)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_insertion_adds_stop(self):
        hgvsc = "NM_999999.1:c.8_9insTT"
        hgvsp_expected = "MOCK:p.(Lys4Ter)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_deletion_no_frameshift(self):
        hgvsc = "NM_999999.1:c.10_12del"
        hgvsp_expected = "MOCK:p.(Lys4del)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_deletion2_no_frameshift(self):
        hgvsc = "NM_999999.1:c.4_15del"
        hgvsp_expected = "MOCK:p.(Lys2_Ala5del)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_deletion3_no_frameshift_c_term(self):
        hgvsc = "NM_999995.1:c.4_6del"
        hgvsp_expected = "MOCK:p.(Lys3del)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_deletion4_no_frameshift_c_term(self):
        hgvsc = "NM_999994.1:c.4_9del"
        hgvsp_expected = "MOCK:p.(Lys3_Lys4del)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_deletion5_no_frameshift(self):
        hgvsc = "NM_999994.1:c.20_25del"
        hgvsp_expected = "MOCK:p.(Ala7_Arg9delinsGly)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_deletion6_no_frameshift(self):
        hgvsc = "NM_999999.1:c.5_7del"
        hgvsp_expected = "MOCK:p.(Lys2_Ala3delinsThr)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_deletion7_no_frameshift(self):
        hgvsc = "NM_999993.1:c.13_24del"
        hgvsp_expected = "MOCK:p.(Arg5_Ala8del)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_deletion_frameshift_nostop(self):
        hgvsc = "NM_999999.1:c.11_12del"
        hgvsp_expected = "MOCK:p.(Lys4SerfsTer?)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_deletion_frameshift_adds_stop(self):
        hgvsc = "NM_999997.1:c.7del"
        hgvsp_expected = "MOCK:p.(Ala3ArgfsTer6)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_deletion_no_frameshift_removes_stop_plus_previous(self):
        hgvsc = "NM_999999.1:c.25_30del"
        hgvsp_expected = "MOCK:p.(Lys9_Ter10delinsGly)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_indel_no_frameshift(self):
        hgvsc = "NM_999999.1:c.11_12delinsTCCCA"
        hgvsp_expected = "MOCK:p.(Lys4delinsIlePro)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_indel2_no_frameshift(self):
        hgvsc = "NM_999999.1:c.11_18delinsTCCCA"
        hgvsp_expected = "MOCK:p.(Lys4_Phe6delinsIlePro)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_indel_frameshift_nostop(self):
        hgvsc = "NM_999999.1:c.8delinsGG"
        hgvsp_expected = "MOCK:p.(Ala3GlyfsTer?)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_dup_1AA_no_frameshift_2(self):
        hgvsc = "NM_999999.1:c.10_12dup"
        hgvsp_expected = "MOCK:p.(Lys4dup)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_dup_1AA_no_frameshift(self):
        hgvsc = "NM_999999.1:c.16_18dup"
        hgvsp_expected = "MOCK:p.(Phe6dup)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_dup_2AA_no_frameshift(self):
        hgvsc = "NM_999999.1:c.16_21dup"
        hgvsp_expected = "MOCK:p.(Phe6_Arg7dup)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_dup_2AA2_no_frameshift(self):
        hgvsc = "NM_999995.1:c.4_6dup"
        hgvsp_expected = "MOCK:p.(Lys3dup)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_dup_3AA_no_frameshift(self):
        hgvsc = "NM_999999.1:c.16_24dup"
        hgvsp_expected = "MOCK:p.(Phe6_Ala8dup)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_dup_frameshift(self):
        hgvsc = "NM_999999.1:c.12_13dup"
        hgvsp_expected = "MOCK:p.(Ala5GlufsTer?)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_intron(self):
        hgvsc = "NM_999999.1:c.12+1G>A"
        hgvsp_expected = "MOCK:p.?"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_five_prime_utr(self):
        hgvsc = "NM_999999.1:c.-2A>G"
        hgvsp_expected = "MOCK:p.?"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_three_prime_utr(self):
        hgvsc = "NM_999999.1:c.*3G>A"
        hgvsp_expected = "MOCK:p.?"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_deletion_into_three_prime_utr_frameshift(self):
        hgvsc = "NM_999999.1:c.27_*3del"
        hgvsp_expected = "MOCK:p.(Lys9XaafsTer?)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_deletion_into_three_prime_utr_no_frameshift(self):
        hgvsc = "NM_999995.1:c.28_*3del"
        hgvsp_expected = "MOCK:p.(Lys10_Ter11delinsArgGlnPheArg)"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_delins_into_three_prime_utr_no_frameshift(self):
        hgvsc = "NM_999995.1:c.28_*3delinsGGG"
        hgvsp_expected = "MOCK:p.(Lys10_Ter11delinsGlyArgGlnPheArg)"
        self._run_conversion(hgvsc, hgvsp_expected)

    # See recommendations re p.? (p.Met1?) at:
    # http://varnomen.hgvs.org/recommendations/protein/variant/substitution/
    def test_substitution_removes_start_codon(self):
        hgvsc = "NM_999999.1:c.1A>G"
        hgvsp_expected = "MOCK:p.Met1?"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_deletion_from_five_prime_utr_frameshift(self):
        hgvsc = "NM_999999.1:c.-3_1del"
        hgvsp_expected = "MOCK:p.Met1?"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_deletion_from_five_prime_utr_no_frameshift(self):
        hgvsc = "NM_999999.1:c.-3_3del"
        hgvsp_expected = "MOCK:p.Met1?"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_delins_from_five_prime_utr_no_frameshift(self):
        hgvsc = "NM_999999.1:c.-3_3delinsAAA"
        hgvsp_expected = "MOCK:p.Met1?"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_delete_entire_gene(self):
        hgvsc = "NM_999999.1:c.-3_*1del"
        hgvsp_expected = "MOCK:p.0?"
        self._run_conversion(hgvsc, hgvsp_expected)

    def test_multiple_stop_codons(self):
        hgvsc = "NM_999992.1:c.4G>A"
        hgvsp_expected = "MOCK:p.?"
        self._run_conversion(hgvsc, hgvsp_expected)

    # The following are unsupported
    #
    # def test_repeats(self):
    #     hgvsc = "NM_999999.1:c.12_13[3]"
    #     hgvsp_expected = ""
    #     self._run_conversion(hgvsc, hgvsp_expected)
    #
    # def test_variable_repeats(self):
    #     pass
    #
    # def test_indeterminate_entire_exon_del(self):
    #     pass
    #
    # def test_indeterminate_entire_exon_dup(self):
    #     pass
    #
    # def test_mosaic(self):
    #     pass
    #
    # def test_chimera(self):
    #     pass
    #
    # def test_two_changes_same_allele(self):
    #     pass
    #
    # def test_two_changes_diff_allele(self):
    #     pass
    #
    # def test_two_changes_unknown_allele(self):
    #     pass

    def _run_conversion(self, hgvsc, hgvsp_expected):
        """Helper method to actually run the test
        :param hgvsc tag
        """
        var_c = TestHgvsCToP._parser.parse_hgvs_variant(hgvsc)
        ac_p = "MOCK"
        hgvsp_actual = str(TestHgvsCToP._mapper.c_to_p(var_c, ac_p))
        msg = "hgvsc: {} hgvsp expected: {} actual: {}".format(hgvsc, hgvsp_expected, hgvsp_actual)
        self.assertEqual(hgvsp_expected, hgvsp_actual, msg)

    # TODO - review other classes of hgvs tags (e.g. utr, intronic) - more use cases?
    # 5'utr
    # intronic
    # after stop codon
    # uncertainties in dups/dels (i.e. hgvs tags with ?)


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os

import unittest

import pytest

from hgvs.exceptions import HGVSError, HGVSUnsupportedOperationError, HGVSInvalidVariantError
import hgvs.dataproviders.uta
import hgvs.variantmapper
import hgvs.parser
import hgvs.normalizer
from support import CACHE

hdp = hgvs.dataproviders.uta.connect(mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE)


@pytest.mark.normalization
class Test_HGVSNormalizer(unittest.TestCase):
    """Tests for normalizer"""

    @classmethod
    def setUpClass(cls):
        cls.hp = hgvs.parser.Parser()
        cls.norm = hgvs.normalizer.Normalizer(hdp, shuffle_direction=3, cross_boundaries=True)
        cls.norm5 = hgvs.normalizer.Normalizer(hdp, shuffle_direction=5, cross_boundaries=True)
        cls.normc = hgvs.normalizer.Normalizer(hdp, shuffle_direction=3, cross_boundaries=False)
        cls.norm5c = hgvs.normalizer.Normalizer(hdp, shuffle_direction=5, cross_boundaries=False)

    def test_c_normalizer(self):
        """Test normalizer for variant type c."""
        #3' shuffling
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_000088.3:c.589_600inv"))),
            "NM_000088.3:c.590_599inv")
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_001166478.1:c.31del"))),
            "NM_001166478.1:c.35del")
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_001166478.1:c.35_36insT"))),
            "NM_001166478.1:c.35dup")
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_001166478.1:c.36_37insTC"))),
            "NM_001166478.1:c.36_37dup")
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_001166478.1:c.35_36dup"))),
            "NM_001166478.1:c.36_37dup")
        self.assertEqual(
            str(
                self.norm.normalize(
                    self.hp.parse_hgvs_variant("NM_001166478.1:c.2_7delinsTTTAGA"))),
            "NM_001166478.1:c.3_4delinsTT")
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_001166478.1:c.30_31insT"))),
            "NM_001166478.1:c.35dup")
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_001166478.1:c.59delG"))),
            "NM_001166478.1:c.61del")
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_001166478.1:c.36_37insTCTCTC"))),
            "NM_001166478.1:c.37_38insCTCTCT")
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_000051.3:c.14_15insT"))),
            "NM_000051.3:c.15dup")

        #5' shuffling
        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_000088.3:c.589_600inv"))),
            "NM_000088.3:c.590_599inv")
        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_001166478.1:c.34del"))),
            "NM_001166478.1:c.31del")
        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_001166478.1:c.35_36insT"))),
            "NM_001166478.1:c.31dup")
        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_001166478.1:c.36_37insTC"))),
            "NM_001166478.1:c.35_36dup")
        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_001166478.1:c.35_36dup"))),
            "NM_001166478.1:c.35_36dup")
        self.assertEqual(
            str(
                self.norm5.normalize(
                    self.hp.parse_hgvs_variant("NM_001166478.1:c.2_7delinsTTTAGA"))),
            "NM_001166478.1:c.3_4delinsTT")
        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_001166478.1:c.30_31insT"))),
            "NM_001166478.1:c.31dup")
        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_001166478.1:c.61delG"))),
            "NM_001166478.1:c.59del")
        self.assertEqual(
            str(
                self.norm5.normalize(
                    self.hp.parse_hgvs_variant("NM_001166478.1:c.36_37insTCTCTC"))),
            "NM_001166478.1:c.34_35insTCTCTC")
        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_000051.3:c.14_15insT"))),
            "NM_000051.3:c.14dup")

        #Around exon-intron boundary
        self.assertEqual(
            str(self.normc.normalize(self.hp.parse_hgvs_variant("NM_001166478.1:c.59delG"))),
            "NM_001166478.1:c.60del")
        self.assertEqual(
            str(self.norm5c.normalize(self.hp.parse_hgvs_variant("NM_001166478.1:c.61delG"))),
            "NM_001166478.1:c.61del")
        self.assertEqual(
            str(self.norm5c.normalize(self.hp.parse_hgvs_variant("NM_001110792.1:c.1030_1035del"))),
            "NM_001110792.1:c.1029_1034del")
        with self.assertRaises(HGVSUnsupportedOperationError):
            self.normc.normalize(self.hp.parse_hgvs_variant("NM_001166478.1:c.59_61del"))

        #UTR variants
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_000051.3:c.-5_-4insA"))),
            "NM_000051.3:c.-3dup")
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_000051.3:c.-4_-3insAC"))),
            "NM_000051.3:c.-3_-2dup")
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_000051.3:c.-2_-1insCA"))),
            "NM_000051.3:c.-1_1dup")

        self.assertEqual(
            str(self.normc.normalize(self.hp.parse_hgvs_variant("NM_000051.3:c.-2_-1insCA"))),
            "NM_000051.3:c.-1_1insAC")

        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_000051.3:c.-4_-3insA"))),
            "NM_000051.3:c.-4dup")
        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_000051.3:c.1_2insCA"))),
            "NM_000051.3:c.-1_1dup")

        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_000051.3:c.*2_*3insT"))),
            "NM_000051.3:c.*4dup")
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_000051.3:c.9170_9171insAT"))),
            "NM_000051.3:c.9171_*1dup")

        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_000051.3:c.*4_*5insT"))),
            "NM_000051.3:c.*3dup")
        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_000051.3:c.9171_*1insA"))),
            "NM_000051.3:c.9171dup")

        with self.assertRaises(HGVSInvalidVariantError):
            self.norm.normalize(self.hp.parse_hgvs_variant("NM_000059.3:c.7790delAAG"))

    def test_g_normalizer(self):
        """Test normalizer for variant type g."""
        #3' shuffling
        self.assertEqual(
            str(
                self.norm.normalize(
                    self.hp.parse_hgvs_variant("NC_000006.11:g.49917122_49917123insA"))),
            "NC_000006.11:g.49917127dup")
        self.assertEqual(
            str(
                self.norm.normalize(
                    self.hp.parse_hgvs_variant("NC_000006.11:g.49917121_49917122insGA"))),
            "NC_000006.11:g.49917122_49917123dup")
        self.assertEqual(
            str(
                self.norm.normalize(
                    self.hp.parse_hgvs_variant("NC_000006.11:g.49917122_49917123dup"))),
            "NC_000006.11:g.49917122_49917123dup")
        self.assertEqual(
            str(
                self.norm.normalize(
                    self.hp.parse_hgvs_variant("NC_000006.11:g.49917122_49917123dupGA"))),
            "NC_000006.11:g.49917122_49917123dup")
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NC_000006.11:g.49917098delC"))),
            "NC_000006.11:g.49917099del")
        self.assertEqual(
            str(
                self.norm.normalize(
                    self.hp.parse_hgvs_variant("NC_000006.11:g.49917151_49917156delinsTCTAAA"))),
            "NC_000006.11:g.49917154_49917155delinsAA")
        self.assertEqual(
            str(
                self.norm.normalize(
                    self.hp.parse_hgvs_variant("NC_000001.10:g.1647893delinsCTTTCTT"))),
            "NC_000001.10:g.1647895_1647900dup")
        self.assertEqual(
            str(
                self.norm.normalize(
                    self.hp.parse_hgvs_variant(
                        "NC_000003.12:g.46709584_46709610del27insAAGAAGAAGAAGAAGAAGAAGAAGAAG"))),
            "NC_000003.12:g.46709584_46709610=")

        #5' shuffling
        self.assertEqual(
            str(
                self.norm5.normalize(
                    self.hp.parse_hgvs_variant("NC_000006.11:g.49917122_49917123insA"))),
            "NC_000006.11:g.49917123dup")
        self.assertEqual(
            str(
                self.norm5.normalize(
                    self.hp.parse_hgvs_variant("NC_000006.11:g.49917121_49917122insGA"))),
            "NC_000006.11:g.49917121_49917122dup")
        self.assertEqual(
            str(
                self.norm5.normalize(
                    self.hp.parse_hgvs_variant("NC_000006.11:g.49917122_49917123dup"))),
            "NC_000006.11:g.49917121_49917122dup")
        self.assertEqual(
            str(
                self.norm5.normalize(
                    self.hp.parse_hgvs_variant("NC_000006.11:g.49917122_49917123dupGA"))),
            "NC_000006.11:g.49917121_49917122dup")
        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NC_000006.11:g.49917099delC"))),
            "NC_000006.11:g.49917098del")
        self.assertEqual(
            str(
                self.norm5.normalize(
                    self.hp.parse_hgvs_variant("NC_000006.11:g.49917151_49917156delinsTCTAAA"))),
            "NC_000006.11:g.49917154_49917155delinsAA")
        self.assertEqual(
            str(
                self.norm5.normalize(
                    self.hp.parse_hgvs_variant(
                        "NC_000003.12:g.46709584_46709610del27insAAGAAGAAGAAGAAGAAGAAGAAGAAG"))),
            "NC_000003.12:g.46709584_46709610=")

        self.assertEqual(
            str(
                self.norm.normalize(
                    self.hp.parse_hgvs_variant("NC_000009.11:g.36233991_36233992delCAinsTG"))),
            "NC_000009.11:g.36233991_36233992inv")
        with self.assertRaises(HGVSInvalidVariantError):
            self.norm.normalize(
                self.hp.parse_hgvs_variant("NG_032871.1:g.32476_53457delinsAATTAAGGTATA"))

        with self.assertRaises(HGVSInvalidVariantError):
            self.norm.normalize(
                self.hp.parse_hgvs_variant("NG_032871.1:g.32476_53457delinsAATTAAGGTATA"))

    def test_norm_var_near_bound(self):
        """Test normalizing variants near the end or start of transcript"""
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.935T>C"))),
            "NM_001001656.1:c.935T>C")
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.945G>C"))),
            "NM_001001656.1:c.945G>C")
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.945dup"))),
            "NM_001001656.1:c.945dup")
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.935_945del"))),
            "NM_001001656.1:c.935_945del")
        with self.assertRaises(HGVSError):
            self.norm.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.946G>C"))
        with self.assertRaises(HGVSError):
            self.norm.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.946dup"))
        with self.assertRaises(HGVSError):
            self.norm.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.935_946del"))

        self.assertEqual(
            str(self.normc.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.935T>C"))),
            "NM_001001656.1:c.935T>C")
        self.assertEqual(
            str(self.normc.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.945G>C"))),
            "NM_001001656.1:c.945G>C")
        self.assertEqual(
            str(self.normc.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.945dup"))),
            "NM_001001656.1:c.945dup")
        self.assertEqual(
            str(self.normc.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.935_945del"))),
            "NM_001001656.1:c.935_945del")
        with self.assertRaises(HGVSError):
            self.normc.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.946G>C"))
        with self.assertRaises(HGVSError):
            self.normc.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.946dup"))
        with self.assertRaises(HGVSError):
            self.normc.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.935_946del"))

        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.1A>T"))),
            "NM_001001656.1:c.1A>T")
        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.1del"))),
            "NM_001001656.1:c.1del")
        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.1dup"))),
            "NM_001001656.1:c.1dup")

        self.assertEqual(
            str(self.norm5c.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.1A>T"))),
            "NM_001001656.1:c.1A>T")
        self.assertEqual(
            str(self.norm5c.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.1del"))),
            "NM_001001656.1:c.1del")
        self.assertEqual(
            str(self.norm5c.normalize(self.hp.parse_hgvs_variant("NM_001001656.1:c.1dup"))),
            "NM_001001656.1:c.1dup")

        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_212556.2:c.1_2insCA"))),
            "NM_212556.2:c.1_2insCA")
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_212556.2:c.2_3insCAT"))),
            "NM_212556.2:c.2_3insCAT")
        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_212556.2:c.1delinsCA"))),
            "NM_212556.2:c.1delinsCA")

        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_212556.2:c.1_2insCA"))),
            "NM_212556.2:c.1delinsACA")
        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_212556.2:c.2_3insCAT"))),
            "NM_212556.2:c.1delinsATCA")
        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_212556.2:c.1delinsCA"))),
            "NM_212556.2:c.1delinsCA")
        self.assertEqual(
            str(self.norm5.normalize(self.hp.parse_hgvs_variant("NM_212556.2:c.1delinsAA"))),
            "NM_212556.2:c.1dup")

        self.assertEqual(
            str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_212556.2:c.1400_1401insAC"))),
            "NM_212556.2:c.1401delinsACA")
        self.assertEqual(
            str(self.normc.normalize(self.hp.parse_hgvs_variant("NM_212556.2:c.1400_1401insAC"))),
            "NM_212556.2:c.1401delinsACA")
        self.assertEqual(
            str(self.normc.normalize(self.hp.parse_hgvs_variant("NM_212556.2:c.1401delinsAA"))),
            "NM_212556.2:c.1401dup")


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
"""hgvs is a package to parse, format, and manipulate biological sequence
variants.  See https://github.com/biocommons/hgvs/ for details.

Example use:

>>> import hgvs.dataproviders.uta
>>> import hgvs.parser
>>> import hgvs.variantmapper

# start with these variants as strings
>>> hgvs_g, hgvs_c = "NC_000007.13:g.36561662C>T", "NM_001637.3:c.1582G>A"

# parse the genomic variant into a Python structure
>>> hp = hgvs.parser.Parser()
>>> var_g = hp.parse_hgvs_variant(hgvs_g)
>>> var_g
SequenceVariant(ac=NC_000007.13, type=g, posedit=36561662C>T, gene=None)

# SequenceVariants are composed of structured objects, e.g.,
>>> var_g.posedit.pos.start
SimplePosition(base=36561662, uncertain=False)

# format by stringification 
>>> str(var_g)
'NC_000007.13:g.36561662C>T'

# initialize the mapper for GRCh37 with splign-based alignments
>>> hdp = hgvs.dataproviders.uta.connect()
>>> am = hgvs.assemblymapper.AssemblyMapper(hdp,
...          assembly_name="GRCh37", alt_aln_method="splign",
...          replace_reference=True)

# identify transcripts that overlap this genomic variant
>>> transcripts = am.relevant_transcripts(var_g)
>>> sorted(transcripts)
['NM_001177506.1', 'NM_001177507.1', 'NM_001637.3']

# map genomic variant to one of these transcripts
>>> var_c = am.g_to_c(var_g, "NM_001637.3")
>>> var_c
SequenceVariant(ac=NM_001637.3, type=c, posedit=1582G>A, gene=None)
>>> str(var_c)
'NM_001637.3:c.1582G>A'

# CDS coordinates use BaseOffsetPosition to support intronic offsets
>>> var_c.posedit.pos.start
BaseOffsetPosition(base=1582, offset=0, datum=Datum.CDS_START, uncertain=False)

"""

from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import pkg_resources
import re
import warnings
import sys

from .config import global_config    # noqa (importing symbol)

_logger = logging.getLogger(__name__)


if sys.version_info < (3, 6):
    _logger.critical("""hgvs on Python 2.7 is no longer supported. The next major version
    of hgvs will require Python 3.6+. See
    https://github.com/biocommons/org/wiki/Migrating-to-Python-3.6.""")


_is_released_version = False
try:
    __version__ = pkg_resources.get_distribution("hgvs").version
    # TODO: match other valid release tags, like .post\d+
    if re.match(r"^\d+\.\d+\.\d+$", __version__) is not None:
        _is_released_version = True
except pkg_resources.DistributionNotFound as e:
    warnings.warn("can't get __version__ because %s package isn't installed" % __package__, Warning)
    __version__ = None


# Enable DeprecationWarnings for the hgvs package
# N.B. The module name is provided as a regexp to the module *path*
warnings.filterwarnings("default", "", DeprecationWarning, r".*\Wlib\W.*\Whgvs\W.*")


_logger.info("hgvs " + __version__ + "; released: " + str(_is_released_version))


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os

import unittest

import pytest

import hgvs
import hgvs.sequencevariant
import hgvs.parser
from support import CACHE


def test_gene_formatting(parser):
    v = parser.parse("NM_01234.5(BOGUS):c.65A>C")
    assert str(v) == "NM_01234.5(BOGUS):c.65A>C"
    v.gene = None
    assert str(v) == "NM_01234.5:c.65A>C"
    

@pytest.mark.quick
@pytest.mark.models
class Test_SequenceVariant(unittest.TestCase):
    def test_SequenceVariant(self):
        var = hgvs.sequencevariant.SequenceVariant(ac="AC", type="B", posedit="1234DE>FG")
        self.assertEqual(str(var), "AC:B.1234DE>FG")

    def test_fill_ref(self):
        hp = hgvs.parser.Parser()
        hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE)

        # fill reference for sequence variants
        var = hp.parse_hgvs_variant("NM_001166478.1:c.31_32del").fill_ref(hdp)
        self.assertEqual(var.format({'max_ref_length': None}), "NM_001166478.1:c.31_32delTT")

        var = hp.parse_hgvs_variant("NM_001166478.1:c.31_32del2").fill_ref(hdp)
        self.assertEqual(var.format({'max_ref_length': None}), "NM_001166478.1:c.31_32delTT")

        var = hp.parse_hgvs_variant("NM_001166478.1:c.2_7delinsTTTAGA").fill_ref(hdp)
        self.assertEqual(
            var.format({'max_ref_length': None}), "NM_001166478.1:c.2_7delTGAAGAinsTTTAGA")

        var = hp.parse_hgvs_variant("NM_001166478.1:c.35_36dup").fill_ref(hdp)
        self.assertEqual(var.format({'max_ref_length': None}), "NM_001166478.1:c.35_36dupTC")

        var = hp.parse_hgvs_variant("NM_001166478.1:c.18_19insACT").fill_ref(hdp)
        self.assertEqual(var.format({'max_ref_length': None}), "NM_001166478.1:c.18_19insACT")

        var = hp.parse_hgvs_variant("NM_001166478.1:c.31=").fill_ref(hdp)
        self.assertEqual(var.format({'max_ref_length': None}), "NM_001166478.1:c.31T=")

    def test_format(self):
        hp = hgvs.parser.Parser()

        # Global default settings
        var = hp.parse_hgvs_variant("NP_001628.1:p.Gly528Arg")
        self.assertEqual(str(var), "NP_001628.1:p.Gly528Arg")
        self.assertEqual(var.format(), "NP_001628.1:p.Gly528Arg")

        # Change global settings
        hgvs.global_config.formatting.p_3_letter = False
        self.assertEqual(str(var), "NP_001628.1:p.G528R")

        # Custom settings
        hgvs.global_config.formatting.p_3_letter = True
        conf = {"p_3_letter": False}
        self.assertEqual(var.format(conf), "NP_001628.1:p.G528R")

        var = hp.parse_hgvs_variant("NP_001628.1:p.Gly528Ter")
        conf = {"p_term_asterisk": True}
        self.assertEqual(var.format(conf), "NP_001628.1:p.Gly528*")
        self.assertEqual(var.format(), "NP_001628.1:p.Gly528Ter")
        conf = {"p_3_letter": False}
        self.assertEqual(var.format(conf), "NP_001628.1:p.G528*")
        self.assertEqual(var.format(), "NP_001628.1:p.Gly528Ter")

        # Remove reference sequence
        var = hp.parse_hgvs_variant("NM_001166478.1:c.31_32delTT")
        self.assertEqual(str(var), "NM_001166478.1:c.31_32del")
        self.assertEqual(var.format(conf={'max_ref_length': 1}), "NM_001166478.1:c.31_32del")
        self.assertEqual(var.format(conf={'max_ref_length': 2}), "NM_001166478.1:c.31_32delTT")
        self.assertEqual(var.format(conf={'max_ref_length': None}), "NM_001166478.1:c.31_32delTT")

        var = hp.parse_hgvs_variant("NM_001166478.1:c.31_32del2")
        self.assertEqual(str(var), "NM_001166478.1:c.31_32del")
        self.assertEqual(var.format(conf={'max_ref_length': None}), "NM_001166478.1:c.31_32del2")

        var = hp.parse_hgvs_variant("NM_001166478.1:c.31_32delTTinsAA")
        self.assertEqual(str(var), "NM_001166478.1:c.31_32delinsAA")
        var = hp.parse_hgvs_variant("NM_001166478.1:c.35_36dupTC")
        self.assertEqual(str(var), "NM_001166478.1:c.35_36dup")
        var = hp.parse_hgvs_variant("NM_001166478.1:c.31T=")
        self.assertEqual(str(var), "NM_001166478.1:c.31=")
        self.assertEqual(var.format(conf={'max_ref_length': None}), "NM_001166478.1:c.31T=")


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import pprint
import unittest

import pytest

from hgvs.exceptions import HGVSParseError
import hgvs.parser


def test_parser_variants_with_gene_names(parser):
    assert parser.parse("NM_01234.5(BOGUS):c.22+1A>T")
    
    with pytest.raises(hgvs.exceptions.HGVSParseError):
        parser.parse("NM_01234.5(1BOGUS):c.22+1A>T")


class Test_Position(unittest.TestCase):
    longMessage = True

    @classmethod
    def setUpClass(cls):
        cls.parser = hgvs.parser.Parser()

    def test_parser_parse_shorthand(self):
        v = "NM_01234.5:c.22+1A>T"
        assert self.parser.parse_hgvs_variant(v) == self.parser.parse(v)

    def test_parser_gauntlet(self):
        fn = os.path.join(os.path.dirname(__file__), "data", "gauntlet")
        for var in open(fn, "r"):
            var = var.strip()
            if var.startswith("#") or var == "":
                continue
            v = self.parser.parse_hgvs_variant(var)
            self.assertEqual(var, v.format(conf={'max_ref_length': None}),
                             "parse-format roundtrip failed:" + pprint.pformat(v.posedit))

    @pytest.mark.quick
    def test_parser_reject(self):
        fn = os.path.join(os.path.dirname(__file__), "data", "reject")
        for var in open(fn, "r"):
            var, msg = var.strip().split("\t")
            if var.startswith("#") or var == "":
                continue
            with self.assertRaises(HGVSParseError):
                self.parser.parse_hgvs_variant(var)
                self.assertTrue(False, msg="expected HGVSParseError: %s (%s)" % (var, msg))

    @pytest.mark.quick
    def test_parser_posedit_special(self):
        # See note in grammar about parsing p.=, p.?, and p.0
        self.assertEqual(str(self.parser.parse_p_posedit("0")), "0")
        self.assertEqual(str(self.parser.parse_p_posedit("0?")), "0?")
        #self.assertEqual( str(self.parser.parse_p_posedit("(0)")), "0?" )

        self.assertIsNone(self.parser.parse_p_posedit("?"))

        self.assertEqual(str(self.parser.parse_p_posedit("=")), "=")
        #self.assertEqual( str(self.parser.parse_p_posedit("=?")), "(=)" )
        self.assertEqual(str(self.parser.parse_p_posedit("(=)")), "(=)")


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import itertools
import re


def build_tx_cigar(exons, strand):
    """builds a single CIGAR string representing an alignment of the
    transcript sequence to a reference sequence, including introns.
    The input exons are expected to be in transcript order, and the
    resulting CIGAR is also in transcript order. 

    >>> build_tx_cigar([], 1) is None
    True
    """
    cigarelem_re = re.compile(r"\d+[=DIMNX]")

    def _reverse_cigar(c):
        return ''.join(reversed(cigarelem_re.findall(c)))

    if len(exons) == 0:
        return None

    # flip orientation of all CIGARs if on - strand
    if strand == -1:
        cigars = [_reverse_cigar(e["cigar"]) for e in exons]
    else:
        cigars = [e["cigar"] for e in exons]

    tx_cigar = [cigars[0]]    # exon 1
    for i in range(1, len(cigars)):    # and intron + exon pairs thereafter
        intron = str(exons[i]["alt_start_i"] - exons[i - 1]["alt_end_i"]) + "N"
        tx_cigar += [intron, cigars[i]]

    tx_cigar_str = "".join(tx_cigar)

    return tx_cigar_str





# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
"""Projects variants between sequences using AlignmentMapper.



"""

from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import logging

from bioutils.sequences import reverse_complement

import hgvs
import hgvs.location
import hgvs.normalizer
import hgvs.posedit
import hgvs.edit
import hgvs.alignmentmapper
import hgvs.utils.altseq_to_hgvsp as altseq_to_hgvsp
import hgvs.utils.altseqbuilder as altseqbuilder
import hgvs.sequencevariant
import hgvs.validator

from hgvs.exceptions import HGVSUnsupportedOperationError, HGVSInvalidVariantError
from hgvs.decorators.lru_cache import lru_cache
from hgvs.enums import PrevalidationLevel
from hgvs.utils.reftranscriptdata import RefTranscriptData

_logger = logging.getLogger(__name__)


class VariantMapper(object):
    r"""Maps SequenceVariant objects between g., n., r., c., and p. representations.

    g⟷{c,n,r} projections are similar in that c, n, and r variants
    may use intronic coordinates. There are two essential differences
    that distinguish the three types:

    * Sequence start: In n and r variants, position 1 is the sequence
      start; in c variants, 1 is the transcription start site.
    * Alphabet: In n and c variants, sequences are DNA; in
      r. variants, sequences are RNA.

    This differences are summarized in this diagram::

      g ----acgtatgcac--gtctagacgt----      ----acgtatgcac--gtctagacgt----      ----acgtatgcac--gtctagacgt----
            \         \/         /              \         \/         /              \         \/         /
      c      acgtATGCACGTCTAGacgt         n      acgtatgcacgtctagacgt         r      acguaugcacgucuagacgu
                 1                               1                                   1
      p          MetHisValTer

    The g excerpt and exon structures are identical. The g⟷n
    transformation, which is the most basic, accounts for the offset
    of the aligned sequences (shown with "1") and the exon structure.
    The g⟷c transformation is akin to g⟷n transformation, but
    requires an addition offset to account for the translation start
    site (c.1).  The CDS in uppercase. The g⟷c transformation is
    akin to g⟷n transformation with a change of alphabet.

    Therefore, this this code uses g⟷n as the core transformation
    between genomic and c, n, and r variants: All c⟷g and r⟷g
    transformations use n⟷g after accounting for the above
    differences. For example, c_to_g accounts for the transcription
    start site offset, then calls n_to_g.

    All methods require and return objects of type
    :class:`hgvs.sequencevariant.SequenceVariant`.

    """

    def __init__(self,
                 hdp,
                 replace_reference=hgvs.global_config.mapping.replace_reference,
                 prevalidation_level=hgvs.global_config.mapping.prevalidation_level,
                 add_gene_symbol=hgvs.global_config.mapping.add_gene_symbol
                 ):
        """
        :param bool replace_reference: replace reference (entails additional network access)
        :param str prevalidation_level: None or Intrinsic or Extrinsic validation before mapping

        """
        self.hdp = hdp
        self.replace_reference = replace_reference
        self.add_gene_symbol = add_gene_symbol
        if prevalidation_level is None:
            self.prevalidation_level = PrevalidationLevel.NONE
        else:
            self.prevalidation_level = PrevalidationLevel[prevalidation_level.upper()]
        if self.prevalidation_level == PrevalidationLevel.NONE:
            self._validator = None
        elif self.prevalidation_level == PrevalidationLevel.INTRINSIC:
            self._validator = hgvs.validator.IntrinsicValidator(strict=False)
        else:
            self._validator = hgvs.validator.Validator(self.hdp, strict=False)
        self.left_normalizer = hgvs.normalizer.Normalizer(hdp, shuffle_direction=5, variantmapper=self)


    # ############################################################################
    # g⟷t
    def g_to_t(self, var_g, tx_ac, alt_aln_method=hgvs.global_config.mapping.alt_aln_method):
        if not (var_g.type == "g"):
            raise HGVSInvalidVariantError("Expected a g. variant; got " + str(var_g))
        if self._validator:
            self._validator.validate(var_g)
        var_g.fill_ref(self.hdp)
        mapper = self._fetch_AlignmentMapper(
            tx_ac=tx_ac, alt_ac=var_g.ac, alt_aln_method=alt_aln_method)
        if mapper.is_coding_transcript:
            var_out = VariantMapper.g_to_c(
                self, var_g=var_g, tx_ac=tx_ac, alt_aln_method=alt_aln_method)
        else:
            var_out = VariantMapper.g_to_n(
                self, var_g=var_g, tx_ac=tx_ac, alt_aln_method=alt_aln_method)
        return var_out

    def t_to_g(self, var_t, alt_ac, alt_aln_method=hgvs.global_config.mapping.alt_aln_method):
        if var_t.type not in "cn":
            raise HGVSInvalidVariantError("Expected a c. or n. variant; got " + str(var_t))
        if self._validator:
            self._validator.validate(var_t)
        var_t.fill_ref(self.hdp)
        mapper = self._fetch_AlignmentMapper(
            tx_ac=var_t.ac, alt_ac=alt_ac, alt_aln_method=alt_aln_method)
        if mapper.is_coding_transcript:
            var_out = VariantMapper.c_to_g(
                self, var_c=var_t, alt_ac=alt_ac, alt_aln_method=alt_aln_method)
        else:
            var_out = VariantMapper.n_to_g(
                self, var_n=var_t, alt_ac=alt_ac, alt_aln_method=alt_aln_method)
        return var_out

    # ############################################################################
    # g⟷n
    def g_to_n(self, var_g, tx_ac, alt_aln_method=hgvs.global_config.mapping.alt_aln_method):
        """Given a parsed g. variant, return a n. variant on the specified
        transcript using the specified alignment method (default is
        "splign" from NCBI).

        :param hgvs.sequencevariant.SequenceVariant var_g: a variant object
        :param str tx_ac: a transcript accession (e.g., NM_012345.6 or ENST012345678)
        :param str alt_aln_method: the alignment method; valid values depend on data source
        :returns: variant object (:class:`hgvs.sequencevariant.SequenceVariant`) using transcript (n.) coordinates
        :raises HGVSInvalidVariantError: if var_g is not of type "g"

        """

        if not (var_g.type == "g"):
            raise HGVSInvalidVariantError("Expected a g. variant; got " + str(var_g))
        if self._validator:
            self._validator.validate(var_g)
        mapper = self._fetch_AlignmentMapper(
            tx_ac=tx_ac, alt_ac=var_g.ac, alt_aln_method=alt_aln_method)

        if (mapper.strand == -1
            and not hgvs.global_config.mapping.strict_bounds
            and not mapper.g_interval_is_inbounds(var_g.posedit.pos)):
            _logger.info("Renormalizing out-of-bounds minus strand variant on genomic sequence")
            var_g = self.left_normalizer.normalize(var_g)

        var_g.fill_ref(self.hdp)
        pos_n = mapper.g_to_n(var_g.posedit.pos)
        if not pos_n.uncertain:
            edit_n = self._convert_edit_check_strand(mapper.strand, var_g.posedit.edit)
            if edit_n.type == 'ins' and pos_n.start.offset == 0 and pos_n.end.offset == 0 and pos_n.end - pos_n.start > 1:
                pos_n.start.base += 1
                pos_n.end.base -= 1
                edit_n.ref = ''
        else:
            # variant at alignment gap
            pos_g = mapper.n_to_g(pos_n)
            edit_n = hgvs.edit.NARefAlt(
                ref='', alt=self._get_altered_sequence(mapper.strand, pos_g, var_g))
        pos_n.uncertain = var_g.posedit.pos.uncertain
        var_n = hgvs.sequencevariant.SequenceVariant(
            ac=tx_ac, type="n", posedit=hgvs.posedit.PosEdit(pos_n, edit_n))
        if self.replace_reference:
            self._replace_reference(var_n)
        if self.add_gene_symbol:
            self._update_gene_symbol(var_n, var_g.gene)
        return var_n

    def n_to_g(self, var_n, alt_ac, alt_aln_method=hgvs.global_config.mapping.alt_aln_method):
        """Given a parsed n. variant, return a g. variant on the specified
        transcript using the specified alignment method (default is
        "splign" from NCBI).

        :param hgvs.sequencevariant.SequenceVariant var_n: a variant object
        :param str alt_ac: a reference sequence accession (e.g., NC_000001.11)
        :param str alt_aln_method: the alignment method; valid values depend on data source
        :returns: variant object (:class:`hgvs.sequencevariant.SequenceVariant`)
        :raises HGVSInvalidVariantError: if var_n is not of type "n"

        """

        if not (var_n.type == "n"):
            raise HGVSInvalidVariantError("Expected a n. variant; got " + str(var_n))
        if self._validator:
            self._validator.validate(var_n)
        var_n.fill_ref(self.hdp)
        mapper = self._fetch_AlignmentMapper(
            tx_ac=var_n.ac, alt_ac=alt_ac, alt_aln_method=alt_aln_method)
        pos_g = mapper.n_to_g(var_n.posedit.pos)
        if not pos_g.uncertain:
            edit_g = self._convert_edit_check_strand(mapper.strand, var_n.posedit.edit)
            if edit_g.type == 'ins' and pos_g.end - pos_g.start > 1:
                pos_g.start.base += 1
                pos_g.end.base -= 1
                edit_g.ref = ''
        else:
            # variant at alignment gap
            pos_n = mapper.g_to_n(pos_g)
            edit_g = hgvs.edit.NARefAlt(
                ref='', alt=self._get_altered_sequence(mapper.strand, pos_n, var_n))
        pos_g.uncertain = var_n.posedit.pos.uncertain
        var_g = hgvs.sequencevariant.SequenceVariant(
            ac=alt_ac, type="g", posedit=hgvs.posedit.PosEdit(pos_g, edit_g))
        if self.replace_reference:
            self._replace_reference(var_g)
        # No gene symbol for g. variants (actually, *should* for NG, but no way to distinguish)
        return var_g

    # ############################################################################
    # g⟷c
    def g_to_c(self, var_g, tx_ac, alt_aln_method=hgvs.global_config.mapping.alt_aln_method):
        """Given a parsed g. variant, return a c. variant on the specified
        transcript using the specified alignment method (default is
        "splign" from NCBI).

        :param hgvs.sequencevariant.SequenceVariant var_g: a variant object
        :param str tx_ac: a transcript accession (e.g., NM_012345.6 or ENST012345678)
        :param str alt_aln_method: the alignment method; valid values depend on data source
        :returns: variant object (:class:`hgvs.sequencevariant.SequenceVariant`) using CDS coordinates
        :raises HGVSInvalidVariantError: if var_g is not of type "g"

        """

        if not (var_g.type == "g"):
            raise HGVSInvalidVariantError("Expected a g. variant; got " + str(var_g))
        if self._validator:
            self._validator.validate(var_g)
        var_g.fill_ref(self.hdp)
        mapper = self._fetch_AlignmentMapper(
            tx_ac=tx_ac, alt_ac=var_g.ac, alt_aln_method=alt_aln_method)
        pos_c = mapper.g_to_c(var_g.posedit.pos)
        if not pos_c.uncertain:
            edit_c = self._convert_edit_check_strand(mapper.strand, var_g.posedit.edit)
            if edit_c.type == 'ins' and pos_c.start.offset == 0 and pos_c.end.offset == 0 and pos_c.end - pos_c.start > 1:
                pos_c.start.base += 1
                pos_c.end.base -= 1
                edit_c.ref = ''
        else:
            # variant at alignment gap
            pos_g = mapper.c_to_g(pos_c)
            edit_c = hgvs.edit.NARefAlt(
                ref='', alt=self._get_altered_sequence(mapper.strand, pos_g, var_g))
        pos_c.uncertain = var_g.posedit.pos.uncertain
        var_c = hgvs.sequencevariant.SequenceVariant(
            ac=tx_ac, type="c", posedit=hgvs.posedit.PosEdit(pos_c, edit_c))
        if self.replace_reference:
            self._replace_reference(var_c)
        if self.add_gene_symbol:
            self._update_gene_symbol(var_c, var_g.gene)
        return var_c

    def c_to_g(self, var_c, alt_ac, alt_aln_method=hgvs.global_config.mapping.alt_aln_method):
        """Given a parsed c. variant, return a g. variant on the specified
        transcript using the specified alignment method (default is
        "splign" from NCBI).

        :param hgvs.sequencevariant.SequenceVariant var_c: a variant object
        :param str alt_ac: a reference sequence accession (e.g., NC_000001.11)
        :param str alt_aln_method: the alignment method; valid values depend on data source
        :returns: variant object (:class:`hgvs.sequencevariant.SequenceVariant`)
        :raises HGVSInvalidVariantError: if var_c is not of type "c"

        """

        if not (var_c.type == "c"):
            raise HGVSInvalidVariantError("Expected a cDNA (c.); got " + str(var_c))
        if self._validator:
            self._validator.validate(var_c)
        var_c.fill_ref(self.hdp)
        mapper = self._fetch_AlignmentMapper(
            tx_ac=var_c.ac, alt_ac=alt_ac, alt_aln_method=alt_aln_method)
        pos_g = mapper.c_to_g(var_c.posedit.pos)
        if not pos_g.uncertain:
            edit_g = self._convert_edit_check_strand(mapper.strand, var_c.posedit.edit)
            if edit_g.type == 'ins' and pos_g.end - pos_g.start > 1:
                pos_g.start.base += 1
                pos_g.end.base -= 1
                edit_g.ref = ''
        else:
            # variant at alignment gap
            var_n = copy.deepcopy(var_c)
            var_n.posedit.pos = mapper.c_to_n(var_c.posedit.pos)
            var_n.type = 'n'
            pos_n = mapper.g_to_n(pos_g)
            edit_g = hgvs.edit.NARefAlt(
                ref='', alt=self._get_altered_sequence(mapper.strand, pos_n, var_n))
        pos_g.uncertain = var_c.posedit.pos.uncertain
        var_g = hgvs.sequencevariant.SequenceVariant(
            ac=alt_ac, type="g", posedit=hgvs.posedit.PosEdit(pos_g, edit_g))
        if self.replace_reference:
            self._replace_reference(var_g)
        return var_g

    # ############################################################################
    # c⟷n
    def c_to_n(self, var_c):
        """Given a parsed c. variant, return a n. variant on the specified
        transcript using the specified alignment method (default is
        "transcript" indicating a self alignment).

        :param hgvs.sequencevariant.SequenceVariant var_c: a variant object
        :returns: variant object (:class:`hgvs.sequencevariant.SequenceVariant`)
        :raises HGVSInvalidVariantError: if var_c is not of type "c"

        """

        if not (var_c.type == "c"):
            raise HGVSInvalidVariantError("Expected a cDNA (c.); got " + str(var_c))
        if self._validator:
            self._validator.validate(var_c)
        var_c.fill_ref(self.hdp)
        mapper = self._fetch_AlignmentMapper(
            tx_ac=var_c.ac, alt_ac=var_c.ac, alt_aln_method="transcript")
        pos_n = mapper.c_to_n(var_c.posedit.pos)
        if (isinstance(var_c.posedit.edit, hgvs.edit.NARefAlt)
                or isinstance(var_c.posedit.edit, hgvs.edit.Dup)
                or isinstance(var_c.posedit.edit, hgvs.edit.Inv)):
            edit_n = copy.deepcopy(var_c.posedit.edit)
        else:
            raise HGVSUnsupportedOperationError(
                "Only NARefAlt/Dup/Inv types are currently implemented")
        var_n = hgvs.sequencevariant.SequenceVariant(
            ac=var_c.ac, type="n", posedit=hgvs.posedit.PosEdit(pos_n, edit_n))
        if self.replace_reference:
            self._replace_reference(var_n)
        if self.add_gene_symbol:
            self._update_gene_symbol(var_n, var_c.gene)
        return var_n

    def n_to_c(self, var_n):
        """Given a parsed n. variant, return a c. variant on the specified
        transcript using the specified alignment method (default is
        "transcript" indicating a self alignment).

        :param hgvs.sequencevariant.SequenceVariant var_n: a variant object
        :returns: variant object (:class:`hgvs.sequencevariant.SequenceVariant`)
        :raises HGVSInvalidVariantError: if var_n is not of type "n"

        """

        if not (var_n.type == "n"):
            raise HGVSInvalidVariantError("Expected n. variant; got " + str(var_n))
        if self._validator:
            self._validator.validate(var_n)
        var_n.fill_ref(self.hdp)
        mapper = self._fetch_AlignmentMapper(
            tx_ac=var_n.ac, alt_ac=var_n.ac, alt_aln_method="transcript")
        pos_c = mapper.n_to_c(var_n.posedit.pos)
        if (isinstance(var_n.posedit.edit, hgvs.edit.NARefAlt)
                or isinstance(var_n.posedit.edit, hgvs.edit.Dup)
                or isinstance(var_n.posedit.edit, hgvs.edit.Inv)):
            edit_c = copy.deepcopy(var_n.posedit.edit)
        else:
            raise HGVSUnsupportedOperationError(
                "Only NARefAlt/Dup/Inv types are currently implemented")
        var_c = hgvs.sequencevariant.SequenceVariant(
            ac=var_n.ac, type="c", posedit=hgvs.posedit.PosEdit(pos_c, edit_c))
        if self.replace_reference:
            self._replace_reference(var_c)
        if self.add_gene_symbol:
            self._update_gene_symbol(var_c, var_n.gene)
        return var_c

    # ############################################################################
    # c ⟶ p
    def c_to_p(self, var_c, pro_ac=None):
        """
        Converts a c. SequenceVariant to a p. SequenceVariant on the specified protein accession
        Author: Rudy Rico

        :param SequenceVariant var_c: hgvsc tag
        :param str pro_ac: protein accession
        :rtype: hgvs.sequencevariant.SequenceVariant

        """

        if not (var_c.type == "c"):
            raise HGVSInvalidVariantError("Expected a cDNA (c.) variant; got " + str(var_c))
        if self._validator:
            self._validator.validate(var_c)
        reference_data = RefTranscriptData(self.hdp, var_c.ac, pro_ac)
        builder = altseqbuilder.AltSeqBuilder(var_c, reference_data)

        # TODO: handle case where you get 2+ alt sequences back;
        # currently get list of 1 element loop structure implemented
        # to handle this, but doesn't really do anything currently.
        all_alt_data = builder.build_altseq()

        var_ps = []
        for alt_data in all_alt_data:
            builder = altseq_to_hgvsp.AltSeqToHgvsp(reference_data, alt_data)
            var_p = builder.build_hgvsp()
            var_ps.append(var_p)

        var_p = var_ps[0]

        if self.add_gene_symbol:
            self._update_gene_symbol(var_p, var_c.gene)

        return var_p

    ############################################################################
    # Internal methods

    def _replace_reference(self, var):
        """fetch reference sequence for variant and update (in-place) if necessary"""

        if var.type not in "cgmnr":
            raise HGVSUnsupportedOperationError("Can only update references for type c, g, m, n, r")

        if var.posedit.edit.type == "ins":
            # insertions have no reference sequence (zero-width), so return as-is
            return var
        if var.posedit.edit.type == "con":
            # conversions have no reference sequence (zero-width), so return as-is
            return var

        pos = var.posedit.pos
        if ((isinstance(pos.start, hgvs.location.BaseOffsetPosition) and pos.start.offset != 0)
                or (isinstance(pos.end, hgvs.location.BaseOffsetPosition) and pos.end.offset != 0)):
            _logger.info("Can't update reference sequence for intronic variant {}".format(var))
            return var

        # For c. variants, we need coords on underlying sequences
        if var.type == "c":
            mapper = self._fetch_AlignmentMapper(
                tx_ac=var.ac, alt_ac=var.ac, alt_aln_method="transcript")
            pos = mapper.c_to_n(var.posedit.pos)
        else:
            pos = var.posedit.pos
        seq = self.hdp.get_seq(var.ac, pos.start.base - 1, pos.end.base)

        edit = var.posedit.edit
        if edit.ref != seq:
            _logger.debug("Replaced reference sequence in {var} with {seq}".format(
                var=var, seq=seq))
            edit.ref = seq

        return var

    @lru_cache(maxsize=hgvs.global_config.lru_cache.maxsize)
    def _fetch_AlignmentMapper(self, tx_ac, alt_ac, alt_aln_method):
        """
        Get a new AlignmentMapper for the given transcript accession (ac),
        possibly caching the result.
        """
        return hgvs.alignmentmapper.AlignmentMapper(
            self.hdp, tx_ac=tx_ac, alt_ac=alt_ac, alt_aln_method=alt_aln_method)

    @staticmethod
    def _convert_edit_check_strand(strand, edit_in):
        """
        Convert an edit from one type to another, based on the stand and type
        """
        if isinstance(edit_in, hgvs.edit.NARefAlt):
            if strand == 1:
                edit_out = copy.deepcopy(edit_in)
            else:
                try:
                    # if smells like an int, do nothing
                    # TODO: should use ref_n, right?
                    int(edit_in.ref)
                    ref = edit_in.ref
                except (ValueError, TypeError):
                    ref = reverse_complement(edit_in.ref)
                edit_out = hgvs.edit.NARefAlt(
                    ref=ref,
                    alt=reverse_complement(edit_in.alt),
                )
        elif isinstance(edit_in, hgvs.edit.Dup):
            if strand == 1:
                edit_out = copy.deepcopy(edit_in)
            else:
                edit_out = hgvs.edit.Dup(ref=reverse_complement(edit_in.ref))
        elif isinstance(edit_in, hgvs.edit.Inv):
            if strand == 1:
                edit_out = copy.deepcopy(edit_in)
            else:
                try:
                    int(edit_in.ref)
                    ref = edit_in.ref
                except (ValueError, TypeError):
                    ref = reverse_complement(edit_in.ref)
                edit_out = hgvs.edit.Inv(ref=ref)
        else:
            raise NotImplementedError("Only NARefAlt/Dup/Inv types are currently implemented")
        return edit_out

    def _get_altered_sequence(self, strand, interval, var):
        seq = list(self.hdp.get_seq(var.ac, interval.start.base - 1, interval.end.base))
        # positions are 0-based and half-open
        pos_start = var.posedit.pos.start.base - interval.start.base
        pos_end = var.posedit.pos.end.base - interval.start.base + 1
        edit = var.posedit.edit

        if edit.type == 'sub':
            seq[pos_start] = edit.alt
        elif edit.type == 'del':
            del seq[pos_start:pos_end]
        elif edit.type == 'ins':
            seq.insert(pos_start + 1, edit.alt)
        elif edit.type == 'delins':
            del seq[pos_start:pos_end]
            seq.insert(pos_start, edit.alt)
        elif edit.type == 'dup':
            seq.insert(pos_end, ''.join(seq[pos_start:pos_end]))
        elif edit.type == 'inv':
            seq[pos_start:pos_end] = list(reverse_complement(''.join(seq[pos_start:pos_end])))
        elif edit.type == 'identity':
            pass
        else:
            raise HGVSUnsupportedOperationError(
                "Getting altered sequence for {type} is unsupported".format(type=edit.type))

        seq = ''.join(seq)
        if strand == -1:
            seq = reverse_complement(seq)
        return seq

    def _update_gene_symbol(self, var, symbol):
        if not symbol:
            symbol = self.hdp.get_tx_identity_info(var.ac).get("hgnc", None)
        var.gene = symbol            
        return var


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



from setuptools import setup, find_packages
from sys import version_info

short_description = "HGVS Parser, Formatter, Mapper, Validator"
with open("docs/description.txt") as f:
    long_description = f.read()

setup(license="Apache License 2.0 (http://www.apache.org/licenses/LICENSE-2.0)",
      long_description=long_description,
      use_scm_version=True,
      zip_safe=True,
      author="HGVS Contributors",
      author_email = 'hgvs-discuss@gmail.com',
      description = short_description.replace("\n", " "),
      name="hgvs",
      package_data={"hgvs": ["_data/*"]},
      packages=find_packages(),
      url="https://github.com/biocommons/hgvs",
      entry_points={
          'console_scripts': [
              'hgvs-shell = hgvs.shell:shell'
          ],
      },
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "Intended Audience :: Developers",
          "Intended Audience :: Healthcare Industry",
          "Intended Audience :: Science/Research",
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: OS Independent",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python",
          "Topic :: Database :: Front-Ends",
          "Topic :: Scientific/Engineering :: Bio-Informatics",
          "Topic :: Scientific/Engineering :: Medical Science Apps.",
      ],
      keywords=[
          "bioinformatics",
          "computational biology",
          "genome variants",
          "genome variation",
          "genomic variants",
          "genomic variation",
          "genomics",
          "hgvs",
      ],
      install_requires=[
          "attrs>=17.4.0",  # https://github.com/biocommons/hgvs/issues/473
          "biocommons.seqrepo<1.0",
          "bioutils>=0.4.0,<1.0",
          "configparser>=3.3.0",
          "ipython",
          "parsley",
          "psycopg2-binary",
          "six",
      ],
      setup_requires=[
          "pytest-runner",
          "setuptools_scm",
          "wheel",
      ],
      tests_require=[
          "pytest>=5.3",
          "pytest-cov>=2.8",
      ],
)


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
"""Defines the abstract data provider interface

"""

from __future__ import absolute_import, division, print_function, unicode_literals

import abc

import hgvs

from ..decorators.lru_cache import lru_cache, LEARN, RUN, VERIFY
from ..utils.PersistentDict import PersistentDict
from six.moves import map
import six


class Interface(six.with_metaclass(abc.ABCMeta, object)):
    """Variant mapping and validation requires access to external data,
    specifically exon structures, transcript alignments, and protein
    accessions.  In order to isolate the hgvs package from the myriad
    choices and tradeoffs, these data are provided through an
    implementation of the (abstract) HGVS Data Provider Interface.

    As of June 2014, the only available data provider implementation
    uses the `Universal Transcript Archive (UTA)`_, a sister project
    that provides access to transcripts and genome-transcript
    alignments.  `Invitae`_ provides a public UTA database instance
    that is used by default; see the `UTA`_ page for instructions on
    installing your own PostgreSQL or SQLite version.  In the future,
    other implementations may be availablefor other data sources.

    Pure virtural class for the HGVS Data Provider Interface.  Every
    data provider implementation should be a subclass (possibly
    indirect) of this class.

    .. _`Universal Transcript Archive (UTA)`: https://github.com/biocommons/uta
    .. _Invitae: http://invitae.com/
    """

    def interface_version(self):
        return 4

    def __init__(self, mode=None, cache=None):
        """
        :param mode: cache mode (None[default lru cache], 'learn', 'run', 'verify')
        :type mode: str
        :param cache: local cache file name
        :type cache: str
        """
        self.mode = None
        if mode == 'learn':
            self.mode = LEARN
        elif mode == 'run':
            self.mode = RUN
        elif mode == 'verify':
            self.mode = VERIFY

        self.cache = None
        if self.mode is not None:
            if self.mode == LEARN:
                self.cache = PersistentDict(cache, flag='c')
            else:
                self.cache = PersistentDict(cache, flag='r')

        self.data_version = lru_cache(
            maxsize=hgvs.global_config.lru_cache.maxsize, mode=self.mode,
            cache=self.cache)(self.data_version)
        self.schema_version = lru_cache(
            maxsize=hgvs.global_config.lru_cache.maxsize, mode=self.mode,
            cache=self.cache)(self.schema_version)
        self.get_acs_for_protein_seq = lru_cache(
            maxsize=hgvs.global_config.lru_cache.maxsize, mode=self.mode,
            cache=self.cache)(self.get_acs_for_protein_seq)
        self.get_gene_info = lru_cache(
            maxsize=hgvs.global_config.lru_cache.maxsize, mode=self.mode,
            cache=self.cache)(self.get_gene_info)
        self.get_pro_ac_for_tx_ac = lru_cache(
            maxsize=hgvs.global_config.lru_cache.maxsize, mode=self.mode,
            cache=self.cache)(self.get_pro_ac_for_tx_ac)
        self.get_seq = lru_cache(
            maxsize=hgvs.global_config.lru_cache.maxsize, mode=self.mode,
            cache=self.cache)(self.get_seq)
        self.get_similar_transcripts = lru_cache(
            maxsize=hgvs.global_config.lru_cache.maxsize, mode=self.mode,
            cache=self.cache)(self.get_similar_transcripts)
        self.get_tx_exons = lru_cache(
            maxsize=hgvs.global_config.lru_cache.maxsize, mode=self.mode,
            cache=self.cache)(self.get_tx_exons)
        self.get_tx_for_gene = lru_cache(
            maxsize=hgvs.global_config.lru_cache.maxsize, mode=self.mode,
            cache=self.cache)(self.get_tx_for_gene)
        self.get_tx_for_region = lru_cache(
            maxsize=hgvs.global_config.lru_cache.maxsize, mode=self.mode,
            cache=self.cache)(self.get_tx_for_region)
        self.get_tx_identity_info = lru_cache(
            maxsize=hgvs.global_config.lru_cache.maxsize, mode=self.mode,
            cache=self.cache)(self.get_tx_identity_info)
        self.get_tx_info = lru_cache(
            maxsize=hgvs.global_config.lru_cache.maxsize, mode=self.mode,
            cache=self.cache)(self.get_tx_info)
        self.get_tx_mapping_options = lru_cache(
            maxsize=hgvs.global_config.lru_cache.maxsize, mode=self.mode,
            cache=self.cache)(self.get_tx_mapping_options)

        def _split_version_string(v):
            versions = list(map(int, v.split(".")))
            if len(versions) < 2:
                versions += [0]
            assert len(versions) == 2
            return versions

        assert self.required_version is not None

        rv = _split_version_string(self.required_version)
        av = _split_version_string(self.schema_version())

        if av[0] == rv[0] and av[1] >= rv[1]:
            return

        raise RuntimeError(
            "Incompatible versions: {k} requires schema version {rv}, but {self.url} provides version {av}"
            .format(
                k=type(self).__name__,
                self=self,
                rv=self.required_version,
                av=self.schema_version()))

    # required_version: what version of the remote schema is required
    # by the subclass? This value is compared to the result of
    # schema_version, which must be implemented by the class.
    required_version = None

    @abc.abstractmethod
    def data_version(self):
        pass

    @abc.abstractmethod
    def schema_version(self):
        pass

    @abc.abstractmethod
    def get_acs_for_protein_seq(self, seq):
        pass

    @abc.abstractmethod
    def get_assembly_map(self, assembly_name):
        pass

    @abc.abstractmethod
    def get_gene_info(self, gene):
        pass

    @abc.abstractmethod
    def get_pro_ac_for_tx_ac(self, tx_ac):
        pass

    @abc.abstractmethod
    def get_seq(self, ac, start_i=None, end_i=None):
        pass

    @abc.abstractmethod
    def get_similar_transcripts(self, tx_ac):
        pass

    @abc.abstractmethod
    def get_tx_exons(self, tx_ac, alt_ac, alt_aln_method):
        pass

    @abc.abstractmethod
    def get_tx_for_gene(self, gene):
        pass

    @abc.abstractmethod
    def get_tx_for_region(self, alt_ac, alt_aln_method, start_i, end_i):
        pass

    @abc.abstractmethod
    def get_tx_identity_info(self, tx_ac):
        pass

    @abc.abstractmethod
    def get_tx_info(self, tx_ac, alt_ac, alt_aln_method):
        pass

    @abc.abstractmethod
    def get_tx_mapping_options(self, tx_ac):
        pass


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
"""implements validation of hgvs variants

"""

from __future__ import absolute_import, division, print_function, unicode_literals

import logging


import hgvs
import hgvs.parser
import hgvs.edit
import hgvs.variantmapper
from hgvs.exceptions import HGVSInvalidVariantError, HGVSUnsupportedOperationError
from hgvs.enums import ValidationLevel, Datum

SEQ_ERROR_MSG = "Variant reference ({var_ref_seq}) does not agree with reference sequence ({ref_seq})"
CDS_BOUND_ERROR_MSG = "Variant is outside CDS bounds (CDS length : {cds_length})"
TX_BOUND_ERROR_MSG = "Variant is outside the transcript bounds"

BASE_OFFSET_COORD_TYPES = "cnr"
SIMPLE_COORD_TYPES = "gmp"


_logger = logging.getLogger(__name__)


class Validator(object):
    """invoke intrinsic and extrinsic validation"""

    def __init__(self, hdp, strict=hgvs.global_config.validator.strict):
        self.strict = strict
        self._ivr = IntrinsicValidator(strict)
        self._evr = ExtrinsicValidator(hdp, strict)

    def validate(self, var, strict=None):
        if strict is None:
            strict = self.strict
        return self._ivr.validate(var, strict) and self._evr.validate(var, strict)


class IntrinsicValidator(object):
    """
    Attempts to determine if the HGVS name is internally consistent

    """

    def __init__(self, strict=hgvs.global_config.validator.strict):
        self.strict = strict

    def validate(self, var, strict=None):
        assert isinstance(var, hgvs.sequencevariant.SequenceVariant
                          ), "variant must be a parsed HGVS sequence variant object"
        if strict is None:
            strict = self.strict
        fail_level = ValidationLevel.WARNING if strict else ValidationLevel.ERROR
        (res, msg) = var.validate()
        if res >= fail_level:
            raise HGVSInvalidVariantError(msg)
        return True


class ExtrinsicValidator():
    """
    Attempts to determine if the HGVS name validates against external data sources
    """

    def __init__(self, hdp, strict=hgvs.global_config.validator.strict):
        self.strict = strict
        self.hdp = hdp
        self.vm = hgvs.variantmapper.VariantMapper(self.hdp, prevalidation_level=None)

    def validate(self, var, strict=None):
        assert isinstance(var, hgvs.sequencevariant.SequenceVariant
                          ), "variant must be a parsed HGVS sequence variant object"
        if strict is None: strict = self.strict
        fail_level = ValidationLevel.WARNING if strict else ValidationLevel.ERROR

        var_n = None
        if var.type == "n":
            var_n = var
        elif var.type == "c":
            var_n = self.vm.c_to_n(var)
        if var_n is not None:
            res, msg = self._n_within_transcript_bounds(var_n)
            if res != ValidationLevel.VALID:
                if hgvs.global_config.mapping.strict_bounds:
                    raise HGVSInvalidVariantError(msg)
                _logger.warning("{}: Variant outside transcript bounds;"
                                " no validation provided".format(var))
                return True         # no other checking performed

        res, msg = self._c_within_cds_bound(var)
        if res >= fail_level:
            raise HGVSInvalidVariantError(msg)

        res, msg = self._ref_is_valid(var)
        if res >= fail_level:
            raise HGVSInvalidVariantError(msg)

        return True

    def _ref_is_valid(self, var):
        # use reference sequence of original variant, even if later converted (eg c_to_n)
        if (var.type in BASE_OFFSET_COORD_TYPES and var.posedit.pos is not None
                and (var.posedit.pos.start.offset != 0 or var.posedit.pos.end.offset != 0)):
            return (ValidationLevel.WARNING,
                    "Cannot validate sequence of an intronic variant ({})".format(str(var)))

        ref_checks = []
        if var.type == 'p':
            if (not var.posedit
                or not var.posedit.pos
                or not var.posedit.pos.start
                or not var.posedit.pos.end):
                return (ValidationLevel.VALID, None)
            ref_checks.append((var.ac, var.posedit.pos.start.pos, var.posedit.pos.start.pos,
                               var.posedit.pos.start.aa))
            if var.posedit.pos.start.pos != var.posedit.pos.end.pos:
                ref_checks.append((var.ac, var.posedit.pos.end.pos, var.posedit.pos.end.pos,
                                   var.posedit.pos.end.aa))
        else:
            var_ref_seq = getattr(var.posedit.edit, "ref", None) or None
            var_n = self.vm.c_to_n(var) if var.type == "c" else var
            ref_checks.append((var_n.ac, var_n.posedit.pos.start.base, var_n.posedit.pos.end.base,
                                   var_ref_seq))

        for ac, var_ref_start, var_ref_end, var_ref_seq in ref_checks:
            if var_ref_start is None or var_ref_end is None or not var_ref_seq:
                continue

            # ref_seq is digit, as in "del6"
            try:
                int(var_ref_seq)
                continue
            except ValueError:
                pass

            ref_seq = self.hdp.get_seq(ac, var_ref_start - 1, var_ref_end)
            if ref_seq != var_ref_seq:
                return (ValidationLevel.ERROR, str(var) + ": " + SEQ_ERROR_MSG.format(
                    ref_seq=ref_seq, var_ref_seq=var_ref_seq))

        return (ValidationLevel.VALID, None)

    def _c_within_cds_bound(self, var):
        if var.type != 'c':
            return (ValidationLevel.VALID, None)
        tx_info = self.hdp.get_tx_identity_info(var.ac)
        if tx_info is None:
            return (ValidationLevel.WARNING,
                    "No transcript data for accession: {ac}".format(ac=var.ac))
        cds_length = tx_info["cds_end_i"] - tx_info["cds_start_i"]
        if var.posedit.pos.start.datum == Datum.CDS_START and var.posedit.pos.start.base > cds_length:
            return (ValidationLevel.ERROR, CDS_BOUND_ERROR_MSG.format(cds_length=cds_length))
        if var.posedit.pos.end.datum == Datum.CDS_START and var.posedit.pos.end.base > cds_length:
            return (ValidationLevel.ERROR, CDS_BOUND_ERROR_MSG.format(cds_length=cds_length))
        return (ValidationLevel.VALID, None)

    def _n_within_transcript_bounds(self, var):
        if var.type != 'n':
            return (ValidationLevel.VALID, None)
        tx_info = self.hdp.get_tx_identity_info(var.ac)
        tx_len = sum(tx_info["lengths"])
        if tx_info is None:
            return (ValidationLevel.WARNING,
                    "No transcript data for accession: {ac}".format(ac=var.ac))
        if var.posedit.pos.start.datum == Datum.SEQ_START and var.posedit.pos.end.base <= 0:
            return (ValidationLevel.ERROR, TX_BOUND_ERROR_MSG.format())
        if var.posedit.pos.end.datum == Datum.SEQ_START and var.posedit.pos.start.base > tx_len:
            return (ValidationLevel.ERROR, TX_BOUND_ERROR_MSG.format())
        return (ValidationLevel.VALID, None)



# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import pprint
import re
import sys
import os

import unittest

import pytest

from hgvs.exceptions import HGVSError, HGVSDataNotAvailableError, HGVSParseError, HGVSInvalidVariantError
from hgvs.enums import Datum
import hgvs.assemblymapper
import hgvs.dataproviders.uta
import hgvs.normalizer
import hgvs.parser
import hgvs.sequencevariant
import hgvs.validator
import hgvs.variantmapper
from support import CACHE


@pytest.mark.issues
class Test_Issues(unittest.TestCase):
    def setUp(self):
        self.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE)
        self.vm = hgvs.variantmapper.VariantMapper(self.hdp, replace_reference=False)
        self.vm_rr = hgvs.variantmapper.VariantMapper(self.hdp, replace_reference=True)
        self.hp = hgvs.parser.Parser()
        self.hn = hgvs.normalizer.Normalizer(self.hdp)
        self.hv = hgvs.validator.IntrinsicValidator()
        self.am37 = hgvs.assemblymapper.AssemblyMapper(
            self.hdp, replace_reference=True, assembly_name='GRCh37', alt_aln_method='splign')
        self.am38 = hgvs.assemblymapper.AssemblyMapper(
            self.hdp, replace_reference=True, assembly_name='GRCh38', alt_aln_method='splign')
        self.vn = hgvs.normalizer.Normalizer(self.hdp, shuffle_direction=3, cross_boundaries=True)

    def test_260_raise_exception_when_mapping_bogus_variant(self):
        v = self.hp.parse_hgvs_variant("NM_000059.3:c.7790delAAG")
        with self.assertRaises(HGVSInvalidVariantError):
            self.am37.c_to_p(v)

    def test_285_partial_palindrome_inversion(self):
        # https://github.com/biocommons/hgvs/issues/285/
        # Inversion mapping is not palindrome aware
        v = self.hp.parse_hgvs_variant("NM_000088.3:c.589_600inv")
        vn = self.vn.normalize(v)
        self.assertEqual(str(vn), "NM_000088.3:c.590_599inv")

    def test_293_parser_attribute_assignment_error(self):
        # https://github.com/biocommons/hgvs/issues/293/
        var = self.hp.parse_hgvs_variant('NG_029146.1:g.6494delG')
        self.vn.normalize(var)    # per issue, should raise error



from bioutils.sequences import translate_cds

from hgvs.exceptions import HGVSDataNotAvailableError


class RefTranscriptData(object):
    def __init__(self, hdp, tx_ac, pro_ac):
        """helper for generating RefTranscriptData from for c_to_p"""
        tx_info = hdp.get_tx_identity_info(tx_ac)
        tx_seq = hdp.get_seq(tx_ac)

        if tx_info is None or tx_seq is None:
            raise HGVSDataNotAvailableError(
                "Missing transcript data for accession: {}".format(tx_ac))

        # use 1-based hgvs coords
        cds_start = tx_info["cds_start_i"] + 1
        cds_stop = tx_info["cds_end_i"]

        # coding sequences that are not divisable by 3 are not yet supported
        tx_seq_to_translate = tx_seq[cds_start - 1:cds_stop]
        if len(tx_seq_to_translate) % 3 != 0:
            raise NotImplementedError(
                "Transcript {} is not supported because its sequence length of {} is not divisible by 3."
                .format(tx_ac, len(tx_seq_to_translate)))

        protein_seq = translate_cds(tx_seq_to_translate)

        if pro_ac is None:
            # get_acs... will always return at least the MD5_ accession
            # TODO: drop get_acs_for_protein_seq; use known mapping or digest (wo/pro ac inference)
            pro_ac = (hdp.get_pro_ac_for_tx_ac(tx_ac)
                      or hdp.get_acs_for_protein_seq(protein_seq)[0])

        self.transcript_sequence = tx_seq
        self.aa_sequence = protein_seq
        self.cds_start = cds_start
        self.cds_stop = cds_stop
        self.protein_accession = pro_ac



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os

import unittest

import pytest

from hgvs.exceptions import HGVSError, HGVSInvalidVariantError
import hgvs.dataproviders.uta
import hgvs.parser
import hgvs.variantmapper
from support import CACHE


def test_add_gene_symbol(am38, parser):
    ags = am38.add_gene_symbol
    var_g = parser.parse("NC_000007.13:g.21940852_21940908del")

    am38.add_gene_symbol = True
    var_t = am38.g_to_t(var_g, "NM_003777.3")
    assert "NM_003777.3(DNAH11):c.13552_*36del" == str(var_t)

    am38.add_gene_symbol = False
    var_t = am38.g_to_t(var_g, "NM_003777.3")
    assert "NM_003777.3:c.13552_*36del" == str(var_t)

    am38.add_gene_symbol = ags


@pytest.mark.quick
class Test_VariantMapper_Exceptions(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE)
        cls.vm = hgvs.variantmapper.VariantMapper(cls.hdp)
        cls.hp = hgvs.parser.Parser()

    def test_gcrp_invalid_input_type(self):
        hgvs_g = "NC_000007.13:g.36561662C>T"
        hgvs_c = "NM_001637.3:c.1582G>A"

        var_g = self.hp.parse_hgvs_variant(hgvs_g)
        var_c = self.hp.parse_hgvs_variant(hgvs_c)

        cases = {
            "gc": (self.vm.g_to_c, (var_c, "NM_001637.3")),
            "gr": (self.vm.g_to_n, (var_c, "NM_001637.3")),
            "gt": (self.vm.g_to_t, (var_c, "NM_001637.3")),
            "rg": (self.vm.n_to_g, (var_c, "NM_001637.3")),
            "cg": (self.vm.c_to_g, (var_g, "NM_001637.3")),
            "tg": (self.vm.t_to_g, (var_g, "NM_001637.3")),
            "cr": (self.vm.c_to_n, (var_g, )),
            "rc": (self.vm.n_to_c, (var_g, )),
            "cp": (self.vm.c_to_p, (var_g, None)),
        }

        failures = []
        for key in cases:
            try:
                func, args = cases[key]
                var_result = func(*args)
                failures.append(key)
            except hgvs.exceptions.HGVSInvalidVariantError:
                pass

        self.assertFalse(failures, "conversions not failing: {}".format(failures))

    def test_gc_invalid_input_nm_accession(self):
        hgvs_g = "NC_000007.13:g.36561662C>T"
        var_g = self.hp.parse_hgvs_variant(hgvs_g)
        with self.assertRaises(hgvs.exceptions.HGVSError):
            var_p = self.vm.c_to_p(var_g, "NM_999999.1")

    def test_undefined_cds(self):
        """Raise exception when requesting mapping to/from c. with non-coding transcript"""
        hgvs_n = "NR_111984.1:n.44G>A"    # legit
        hgvs_c = "NR_111984.1:c.44G>A"    # bogus: c. with non-coding tx accession
        var_n = self.hp.parse_hgvs_variant(hgvs_n)
        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        tx_ac = var_n.ac

        with self.assertRaises(hgvs.exceptions.HGVSUsageError):
            var_c = self.vm.n_to_c(var_n)    # n_to_c: transcript is non-coding

        with self.assertRaises(hgvs.exceptions.HGVSUsageError):
            var_c = self.vm.c_to_n(var_c)    # c_to_n: var_c is bogus

    def test_map_var_of_unsupported_validation(self):
        hgvs_c = "NM_003777.3:c.13552_*36del57"
        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_g = self.vm.c_to_g(var_c, "NC_000007.13")
        self.assertEqual(str(var_g), "NC_000007.13:g.21940852_21940908del")

    def test_map_to_unknown_p_effect(self):
        hgvs_c = "NM_020975.4:c.625+9C>T"
        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_p = self.vm.c_to_p(var_c)
        self.assertEqual(str(var_p), "NP_066124.1:p.?")

    def test_map_of_c_out_of_cds_bound(self):
        hgvs_c = "NM_145901.2:c.343T>C"
        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        with self.assertRaises(HGVSInvalidVariantError):
            self.vm.c_to_p(var_c)

    def test_map_of_dup_at_cds_end(self):
        hgvs_c = "NM_001051.2:c.1257dupG"
        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_p = self.vm.c_to_p(var_c)
        self.assertEqual(str(var_p), "NP_001042.1:p.(=)")

    def test_map_of_c_out_of_reference_bound(self):
        hgvs_c = "NM_000249.3:c.-73960_*46597del"
        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        with pytest.raises(HGVSError, match='coordinate is outside the bounds'):
            self.vm.c_to_p(var_c)


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>
